"""Test dispersion code."""

import csv
from pathlib import Path

import numpy as np
from numpy.testing import assert_almost_equal
import pytest

from morfeus import Dispersion

DATA_DIR = Path(__file__).parent / "data" / "dispersion"


def test_H():
    """Test H atom."""
    disp = Dispersion(
        ["H"], [[0.0, 0.0, 0.0]], radii=[1.54], compute_coefficients=False
    )
    disp._c_n_coefficients = {
        6: np.array([7.6]),
        8: np.array([92.1]),
    }
    disp.compute_p_int()
    assert_almost_equal(disp.p_int, 6.2, decimal=1)


def pytest_generate_tests(metafunc):
    """Generate test data from csv file."""
    if "disp_data" in metafunc.fixturenames:
        with open(DATA_DIR / "periodic_table.csv") as f:
            reader = csv.DictReader(f)
            records = list(reader)
        metafunc.parametrize("disp_data", records)


@pytest.mark.benchmark
def test_reference(disp_data):
    """Test against dispersion reference data."""
    data = disp_data
    p_int_ref = float(data["p_int"])
    elements = [data["symbol"]]
    coordinates = [[0.0, 0.0, 0.0]]
    radii = [float(data["radius"])]
    disp = Dispersion(elements, coordinates, radii=radii, compute_coefficients=False)
    disp._c_n_coefficients = {
        6: np.array([float(data["c6"])]),
        8: np.array([float(data["c8"])]),
    }
    disp.compute_p_int()
    assert_almost_equal(disp.p_int, p_int_ref, decimal=1)



"""Test buried volume code."""

import csv
from pathlib import Path

from numpy.testing import assert_almost_equal
import pytest

from morfeus import BuriedVolume, read_xyz

DATA_DIR = Path(__file__).parent / "data" / "buried_volume"


def test_one():
    """Test one complex."""
    elements, coordinates = read_xyz(DATA_DIR / "xyz" / "1.xyz")
    bv = BuriedVolume(elements, coordinates, 1, excluded_atoms=[1, 2, 3, 4, 5, 6, 7])
    assert_almost_equal(bv.fraction_buried_volume * 100, 29.6, decimal=1)


def pytest_generate_tests(metafunc):
    """Generate test data from csv file."""
    if "bv_data" in metafunc.fixturenames:
        with open(DATA_DIR / "reference_data.csv") as f:
            reader = csv.DictReader(f)
            records = list(reader)
        metafunc.parametrize("bv_data", records)


@pytest.mark.benchmark
def test_reference(bv_data):
    """Test against buried volume reference data."""
    data = bv_data
    idx = int(data["idx"])
    percent_buried_volume_ref = float(data["buried_volume"])
    excluded_atoms = [int(i) for i in data["excluded_atoms"].split()]
    elements, coordinates = read_xyz(DATA_DIR / "xyz" / f"{idx}.xyz")
    bv = BuriedVolume(elements, coordinates, 1, excluded_atoms=excluded_atoms)

    assert_almost_equal(
        bv.fraction_buried_volume * 100, percent_buried_volume_ref, decimal=0
    )



"""Test solid angle code."""

import csv
from pathlib import Path

from numpy.testing import assert_almost_equal
import pytest

from morfeus import read_xyz, SolidAngle

DATA_DIR = Path(__file__).parent / "data" / "solid_angle"


def test_PdPMe3():
    """Test PdPMe3."""
    elements, coordinates = read_xyz(DATA_DIR / "pd/PdPMe3.xyz")
    sa = SolidAngle(elements, coordinates, 1, radii_type="bondi")
    assert_almost_equal(sa.solid_angle, 2.877080666, decimal=1)
    assert_almost_equal(sa.cone_angle, 114.3468001, decimal=0)


def pytest_generate_tests(metafunc):
    """Generate test data from csv file."""
    if "solid_angle_data" in metafunc.fixturenames:
        with open(DATA_DIR / "solid_angles.csv") as f:
            reader = csv.DictReader(f)
            records = [("standard", record) for record in reader]
        with open(DATA_DIR / "solid_angles_max.csv") as f:
            reader = csv.DictReader(f)
            records += [("maximum", record) for record in reader]
        metafunc.parametrize("solid_angle_data", records)


@pytest.mark.benchmark
def test_reference(solid_angle_data):
    """Test against solid angle reference data."""
    for metal in ("pd", "pt", "ni"):
        label, data = solid_angle_data
        solid_angle_ref = float(data[f"{metal}_solid_angle"])
        cone_angle_ref = float(data[f"{metal}_cone_angle"])
        if label == "standard":
            xyz_path = DATA_DIR / f"{metal}/{data[f'{metal}_xyz']}.xyz"
        elif label == "maximum":
            xyz_path = DATA_DIR / f"{metal}/maximum/{data[f'{metal}_xyz']}.xyz"
        elements, coordinates = read_xyz(xyz_path)
        sa = SolidAngle(elements, coordinates, 1, radii_type="bondi")
        assert_almost_equal(sa.solid_angle, solid_angle_ref, decimal=1)
        assert_almost_equal(sa.cone_angle, cone_angle_ref, decimal=0)



"""Bite angle code."""

from __future__ import annotations

from collections.abc import Iterable
import functools
from typing import Any, cast

import numpy as np

from morfeus.io import read_geometry
from morfeus.typing import Array1DFloat, Array2DFloat, ArrayLike1D, ArrayLike2D


class BiteAngle:
    """Performs and stores the results of a bite angle calculation.

    The bite angle can become 'inverted' if it goes beyond 180°. This cannot be detected
    automatically, so the user has to supply a reference vector that should point in the
    'direction' of the ligand. As a convenience, the vector can be constructed
    automatically from the metal atom to the geometric mean of a set of 'ref_atoms'.

    Args:
        coordinates: Coordinates (Å)
        metal_index: Index of metal atom (1-indexed)
        ligand_index_1: Index of first ligand atom (1-indexed)
        ligand_index_2: Index of second ligand atom (1-indexed)
        ref_vector: Reference vector for determining inverted bite angle
        ref_atoms: Reference atoms for determnining inverted bite angle (1-indexed)

    Attributes:
        angle: Bite angle (degrees)
        inverted: Whether angle is 'inverted'
    """

    angle: float
    inverted: bool

    def __init__(
        self,
        coordinates: ArrayLike2D,
        metal_index: int,
        ligand_index_1: int,
        ligand_index_2: int,
        ref_atoms: Iterable[int] | None = None,
        ref_vector: ArrayLike1D | None = None,
    ) -> None:
        if 0 in {
            metal_index,
            ligand_index_1,
            ligand_index_2,
            *(ref_atoms or ()),
        }:
            raise IndexError("Atom indices should not be 0 (1-indexed).")

        # Check keywords
        if ref_atoms is not None and ref_vector is not None:
            raise ValueError("ref_atoms and ref_vector cannot be set at the same time.")

        coordinates: Array2DFloat = np.asarray(coordinates)

        # Construct vectors
        v_1 = coordinates[ligand_index_1 - 1] - coordinates[metal_index - 1]
        v_1_norm = v_1 / np.linalg.norm(v_1)
        v_2 = coordinates[ligand_index_2 - 1] - coordinates[metal_index - 1]
        v_2_norm = v_2 / np.linalg.norm(v_2)

        # Calculate angle between vectors
        angle_rad = np.arctan2(
            np.linalg.norm(np.cross(v_1_norm, v_2_norm)), np.dot(v_1_norm, v_2_norm)
        )
        angle = np.rad2deg(angle_rad)

        # Check if angle should be inverted
        if ref_atoms is not None:
            ref_vector = (
                np.mean(coordinates[[i - 1 for i in ref_atoms]], axis=0)
                - coordinates[metal_index - 1]
            )
        ref_vector = cast(Array1DFloat, ref_vector)
        inverted = False
        if ref_vector is not None:
            ref_vector /= np.linalg.norm(ref_vector)
            v_mean = (v_1 + v_2) / 2
            v_mean /= np.linalg.norm(v_mean)
            if np.dot(v_mean, ref_vector) < 0:
                inverted = True
                angle = 360 - angle

        self.angle = angle
        self.inverted = inverted


def cli(file: str) -> Any:
    """CLI for bite angle.

    Args:
        file: Geometry file

    Returns:
        Partially instantiated class
    """
    _, coordinates = read_geometry(file)
    return functools.partial(BiteAngle, coordinates)



"""Test buried volume code."""

import csv
from pathlib import Path

from numpy.testing import assert_almost_equal
import pytest

from morfeus import BiteAngle, read_xyz

DATA_DIR = Path(__file__).parent / "data" / "bite_angle"


def test_one():
    """Test one complex."""
    _, coordinates = read_xyz(DATA_DIR / "xyz" / "cis-B01_XantPhos.xyz")
    bite_angle = BiteAngle(coordinates, 60, 1, 37)
    assert_almost_equal(bite_angle.angle, 101.91, decimal=2)


def test_inverted_vector():
    """Test inverted bite angle by reference vector method."""
    _, coordinates = read_xyz(DATA_DIR / "xyz" / "inverted.xyz")
    ref_vector = coordinates[5] - coordinates[10]
    bite_angle = BiteAngle(coordinates, 11, 10, 14, ref_vector=ref_vector)
    assert_almost_equal(bite_angle.angle, 187.33, decimal=2)


def test_inverted_atoms():
    """Test inverted bite angle by reference vector method."""
    _, coordinates = read_xyz(DATA_DIR / "xyz" / "inverted.xyz")
    bite_angle = BiteAngle(coordinates, 11, 10, 14, ref_atoms=[6])
    assert_almost_equal(bite_angle.angle, 187.33, decimal=2)


def pytest_generate_tests(metafunc):
    """Generate test data from csv file."""
    if "bite_angle_data" in metafunc.fixturenames:
        with open(DATA_DIR / "reference_data.csv") as f:
            reader = csv.DictReader(f)
            records = list(reader)
        metafunc.parametrize("bite_angle_data", records)


@pytest.mark.benchmark
def test_reference(bite_angle_data):
    """Test against buried volume reference data."""
    data = bite_angle_data
    metal_idx = int(data["idx_metal"])
    ligand_idx_1 = int(data["idx_ligand_1"])
    ligand_idx_2 = int(data["idx_ligand_2"])
    bite_angle_ref = float(data["bite_angle"])
    filename = data["filename"]
    _, coordinates = read_xyz(DATA_DIR / "xyz" / filename)
    bite_angle = BiteAngle(coordinates, metal_idx, ligand_idx_1, ligand_idx_2)

    assert_almost_equal(bite_angle.angle, bite_angle_ref, decimal=2)



"""Solid angle code."""

from __future__ import annotations

from collections.abc import Iterable
import functools
import typing
from typing import Any

import numpy as np

from morfeus.data import jmol_colors
from morfeus.geometry import Atom, Sphere
from morfeus.io import read_geometry
from morfeus.typing import Array1DFloat, Array2DFloat, ArrayLike1D, ArrayLike2D
from morfeus.utils import convert_elements, get_radii, Import, requires_dependency

if typing.TYPE_CHECKING:
    from matplotlib.colors import hex2color
    import pyvista as pv
    from pyvistaqt import BackgroundPlotter


class SolidAngle:
    """Calculates and stores the results of a solid angle calculation.

    As described in J. Chem. Theory Comput. 2013, 9 (12), 5734-5744 and Dalton Trans.
    2006, 33, 3991.

    Args:
        elements: Elements as atomic symbols or numbers
        coordinates: Coordinates (Å)
        metal_index: Index of central atom (1-indexed)
        radii: vdW radii (Å)
        radii_type: Type of vdW radii: 'alvarez', 'bondi', 'crc' or 'truhlar'
        density: Area per point (Å²) on the sphere surface

    Attributes:
        cone_angle: Solid cone angle (degrees)
        solid_angle: Solid angle (steradians)
        G: G parameter
    """

    G: float
    cone_angle: float
    solid_angle: float
    _atoms: list[Atom]

    def __init__(
        self,
        elements: Iterable[int] | Iterable[str],
        coordinates: ArrayLike2D,
        metal_index: int,
        radii: ArrayLike1D | None = None,
        radii_type: str = "crc",
        density: float = 0.001,
    ) -> None:
        if metal_index == 0:
            raise IndexError("Atom indices should not be 0 (1-indexed).")

        # Convert elements to atomic numbers if they are symbols
        elements = convert_elements(elements, output="numbers")
        coordinates: Array2DFloat = np.array(coordinates)

        # Get radii if they are not supplied
        if radii is None:
            radii = get_radii(elements, radii_type=radii_type)
        radii: Array1DFloat = np.asarray(radii)
        coordinates_metal = coordinates[metal_index - 1, :]
        coordinates -= coordinates_metal

        # Get list of atoms as Atom objects
        atoms = []
        for i, (element, coords, radius) in enumerate(
            zip(elements, coordinates, radii), start=1
        ):
            if i == metal_index:
                continue
            atom = Atom(element, coords, radius, i)
            atoms.append(atom)

        # Construct sphere and check which points on sphere are within atom cones
        sphere = Sphere(coordinates_metal, 1.0, density=density)

        mask = np.zeros(len(sphere.points), dtype=bool)
        for atom in atoms:
            atom.get_cone()
            is_inside = atom.cone.is_inside_points(sphere.points)
            mask = np.logical_or(mask, is_inside)

        # Calculate solid angle, cone angle and G
        n_occluded = sum(mask)
        n_total = len(mask)
        ratio_occluded = n_occluded / n_total
        area_occluded = ratio_occluded * sphere.area
        omega = area_occluded / sphere.radius**2
        theta = 2 * np.arccos(1 - omega / 2 / np.pi)
        G = 100 * omega / 4 / np.pi

        self.solid_angle = omega
        self.cone_angle = np.rad2deg(theta)
        self.G = G
        self._atoms = atoms
        self._sphere = sphere
        self._mask = mask

    def print_report(self) -> None:
        """Prints report of results."""
        print(f"Solid angle (sr): {self.solid_angle:.3f}")
        print(f"Cone angle (°): {self.cone_angle:.3f}")
        print(f"G: {self.G:.3f}")

    @requires_dependency(
        [
            Import(module="matplotlib.colors", item="hex2color"),
            Import(module="pyvista", alias="pv"),
            Import(module="pyvistaqt", item="BackgroundPlotter"),
        ],
        globals(),
    )
    def draw_3D(
        self,
        atom_scale: float = 1.0,
        background_color: str = "white",
        point_color: str = "steelblue",
        opacity: float = 0.25,
        size: float = 2.0,
    ) -> None:
        """Draw a 3D representation.

        Draws the molecule with the ligand shadow

        Args:
            atom_scale: Scaling factor for atom size
            background_color: Background color for plot
            point_color: Color of surface points
            opacity: Point opacity
            size: Point size
        """
        # Set up plotter
        p = BackgroundPlotter()
        p.set_background(background_color)

        # Draw molecule
        for atom in self._atoms:
            color = hex2color(jmol_colors[atom.element])
            radius = atom.radius * atom_scale
            sphere = pv.Sphere(center=list(atom.coordinates), radius=radius)
            p.add_mesh(sphere, color=color, opacity=1, name=str(atom.index))

        coordinates: Array2DFloat = np.vstack(
            [atom.coordinates for atom in self._atoms]
        )
        radii: Array1DFloat = np.vstack([atom.radius for atom in self._atoms])
        distances = np.linalg.norm(coordinates, axis=1) + radii
        dist_max = np.max(distances)
        sphere_points = self._sphere.points[self._mask] * (dist_max + 0.1)

        # Draw surface points
        p.add_points(
            sphere_points,
            color=point_color,
            opacity=opacity,
            point_size=size,
        )

        sphere = pv.Sphere(center=[0.0, 0.0, 0.0], radius=0.5)
        p.add_mesh(sphere, color="lightsteelblue", opacity=0.75, name=str(1))

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({len(self._atoms)!r} atoms)"


def cli(file: str) -> Any:
    """CLI for solid angle.

    Args:
        file: Geometry file

    Returns:
        Partially instantiated class
    """
    elements, coordinates = read_geometry(file)
    return functools.partial(SolidAngle, elements, coordinates)



"""Test local force code."""

import csv
from pathlib import Path

from numpy.testing import assert_almost_equal
import pytest

from morfeus import LocalForce

DATA_DIR = Path(__file__).parent / "data" / "local_force"


def test_one():
    """Test molecule one."""
    lf = LocalForce()
    lf.load_file(DATA_DIR / "1" / "freq-hp.log", "gaussian", "log")
    lf.compute_local()
    lf.compute_frequencies()
    fc = lf.get_local_force_constant([1, 2])
    freq = lf.get_local_frequency([1, 2])
    assert_almost_equal(fc, 5.365, decimal=1)
    assert_almost_equal(freq, 3129, decimal=-2)


def pytest_generate_tests(metafunc):
    """Generate test data from csv file."""
    if "lf_data" in metafunc.fixturenames:
        with open(DATA_DIR / "reference_data.csv") as f:
            reader = csv.DictReader(f)
            records = list(reader)
        metafunc.parametrize("lf_data", records)


@pytest.mark.benchmark
def test_reference(lf_data):
    """Test against local force reference data."""
    # Test criteria are relatively loose as reference geometries don't exist
    data = lf_data

    fc_ref = float(data["force_constant"])
    freq_ref = float(data["frequency"])
    atoms = [int(i) for i in data["atoms"].split()]
    idx = data["label"].split(".")[0]
    for file_type in ["hp", "lm"]:
        lf = LocalForce()
        lf.load_file(DATA_DIR / f"{idx}" / f"freq-{file_type}.log", "gaussian", "log")
        lf.compute_local()
        lf.compute_frequencies()
        fc = lf.get_local_force_constant(atoms)
        freq = lf.get_local_frequency(atoms)

        assert_almost_equal(fc, fc_ref, decimal=1)
        assert_almost_equal(freq, freq_ref, decimal=-2)



"""Input and output."""

from __future__ import annotations

from collections.abc import Iterable, MutableMapping, Sequence
from os import PathLike
from pathlib import Path
import typing
from typing import Any

import numpy as np

from morfeus.d3_data import r2_r4
from morfeus.data import BOHR_TO_ANGSTROM, KCAL_TO_HARTREE
from morfeus.typing import (
    Array1DAny,
    Array1DFloat,
    Array1DInt,
    Array1DStr,
    Array2DFloat,
    Array2DInt,
    Array3DFloat,
    ArrayLike2D,
    ArrayLike3D,
)
from morfeus.utils import convert_elements, Import, requires_dependency

if typing.TYPE_CHECKING:
    import cclib.io


class CrestParser:
    """Parses folder with CREST output files for conformer information.

    Reads the files: cre_members, crest.energies, crest_conformers.xyz

    Args:
        path: Path to CREST folder

    Attributes:
        conformer_coordinates: Conformer coordinates (Å)
        degeneracies: Degeneracies
        elements: Elements
        energies: Energies (a.u.)
    """

    conformer_coordinates: Array3DFloat
    degeneracies: Array1DInt
    elements: Array1DInt
    energies: Array1DFloat

    def __init__(self, path: str | PathLike) -> None:
        path = Path(path)

        with open(path / "cre_members") as f:
            lines = f.readlines()
        degeneracies = []
        for line in lines:
            strip_line = line.strip().split()
            if len(strip_line) != 1:
                degeneracies.append(int(strip_line[0]))
        degeneracies: Array1DInt = np.array(degeneracies)

        with open(path / "crest.energies") as f:
            lines = f.readlines()

        energies = []
        for line in lines:
            strip_line = line.strip().split()
            energies.append(float(strip_line[1]))
        energies = np.array(energies) * KCAL_TO_HARTREE

        elements, conformer_coordinates = read_xyz(path / "crest_conformers.xyz")
        if conformer_coordinates.ndim == 2:
            conformer_coordinates = conformer_coordinates.reshape(1, -1, 3)

        self.elements = elements
        self.conformer_coordinates = conformer_coordinates
        self.energies = energies
        self.degeneracies = degeneracies


class CubeParser:
    """Parses Gaussian cube file of electron density.

    Args:
        file: Cube file

    Attributes:
        min_x: Minimum x value (Å)
        min_y: Minimum y value (Å)
        min_z: Minimum z value (Å)
        step_x: Step size in x direction (Å)
        step_y: Step size in y direction (Å)
        step_z: Step size in z direction (Å)
        X: 3D array of x values (Å)
        Y: 3D array of y values (Å)
        Z: 3D array of z values (Å)
        S: 3D array of electron density scalars (electorns / Bohr^3)
        n_points_x: Number of points in the x direction
        n_points_y: Number of points in the y direction
        n_points_z: Number of points in the < direction
    """

    min_x: float
    min_y: float
    min_z: float
    step_x: float
    step_y: float
    step_z: float
    X: Array3DFloat
    Y: Array3DFloat
    Z: Array3DFloat
    S: Array3DFloat
    n_points_x: int
    n_points_y: int
    n_points_z: int

    def __init__(self, file: str | PathLike) -> None:
        # Read the lines from the cube file
        with open(file) as file:
            lines = file.readlines()

        # Skip first two lines which are comments
        lines = lines[2:]

        # Get the number of atoms
        n_atoms = int(lines[0].strip().split()[0])

        # Get the minimum values along the axes
        min_x = float(lines[0].strip().split()[1]) * BOHR_TO_ANGSTROM
        min_y = float(lines[0].strip().split()[2]) * BOHR_TO_ANGSTROM
        min_z = float(lines[0].strip().split()[3]) * BOHR_TO_ANGSTROM

        # Get the number of points and step size along each axis
        n_points_x = int(lines[1].strip().split()[0])
        step_x = float(lines[1].strip().split()[1]) * BOHR_TO_ANGSTROM

        n_points_y = int(lines[2].strip().split()[0])
        step_y = float(lines[2].strip().split()[2]) * BOHR_TO_ANGSTROM

        n_points_z = int(lines[3].strip().split()[0])
        step_z = float(lines[3].strip().split()[3]) * BOHR_TO_ANGSTROM

        # Generate grid
        x = min_x + np.arange(0, n_points_x) * step_x
        y = min_y + np.arange(0, n_points_y) * step_y
        z = min_z + np.arange(0, n_points_z) * step_z
        X, Y, Z = np.meshgrid(x, y, z, indexing="ij")

        # Skips to data lines and read in data
        lines = lines[(n_atoms + 4) :]

        data = []
        for line in lines:
            line_data = [float(datum) for datum in line.strip().split()]
            data.extend(line_data)

        # Create array
        S: Array2DFloat = np.array(data).reshape(X.shape)

        # Set up attributes
        self.X = X
        self.Y = Y
        self.Z = Z
        self.S = S

        self.min_x = min_x
        self.min_y = min_y
        self.min_z = min_z

        self.step_x = step_x
        self.step_y = step_y
        self.step_z = step_z

        self.n_points_x = n_points_x
        self.n_points_y = n_points_y
        self.n_points_z = n_points_z

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.S.size!r} points)"


class D3Parser:
    """Parses the output of Grimme's D3 program.

    Extracts the C₆ᴬᴬ and C₈ᴬᴬ coefficients

    Args:
        file: File containing output from the D3 program.

    Attributes:
        c6_coefficients: C₆ᴬᴬ coefficients (a.u.)
        c8_coefficients: C₈ᴬᴬ coefficients (a.u.)
    """

    c6_coefficients: Array1DFloat
    c8_coefficients: Array1DFloat

    def __init__(self, file: str | PathLike) -> None:
        # Read the file
        with open(file, encoding="utf-8") as file:
            lines = file.readlines()

        # Parse the file for the coefficients
        c6_coefficients = []
        c8_coefficients = []
        read = False
        for line in lines:
            if read:
                if not line.strip():
                    read = False
                    break
                strip_line = line.strip().split()
                c6 = float(strip_line[7])
                c8 = float(strip_line[8])
                c6_coefficients.append(c6)
                c8_coefficients.append(c8)
            if "C8(AA)" in line:
                read = True

        # Set attributes
        self.c6_coefficients = np.array(c6_coefficients)
        self.c8_coefficients = np.array(c8_coefficients)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}"


class D4Parser:
    """Parses the output of Grimme's D4 program.

    Extracts the C₆ᴬᴬ and C₈ᴬᴬ coefficients

    Args:
        file: File containing output from the D4 program.

    Attributes:
        c6_coefficients: C₆ᴬᴬ coefficients (a.u.)
        c8_coefficients: C₈ᴬᴬ coefficients (a.u.)
    """

    c6_coefficients: Array1DFloat
    c8_coefficients: Array1DFloat

    def __init__(self, file: str | PathLike) -> None:
        # Read the file
        with open(file, encoding="utf-8") as file:
            lines = file.readlines()

        # Parse the file and extract the coefficients
        c6_coefficients = []
        elements = []
        read = False
        for line in lines:
            if read:
                if not line.strip():
                    read = False
                    break
                strip_line = line.strip().split()
                element = int(strip_line[1])
                elements.append(element)
                c6 = float(strip_line[5])
                c6_coefficients.append(c6)
            if "C6AA" in line:
                read = True
        c6_coefficients: Array1DFloat = np.array(c6_coefficients)
        c8_coefficients: Array1DFloat = np.array(
            [
                3 * c6 * r2_r4[element] ** 2
                for c6, element in zip(c6_coefficients, elements)
            ]
        )

        # Set up attributes
        self.c6_coefficients = c6_coefficients
        self.c8_coefficients = c8_coefficients

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}"


class VertexParser:
    """Parses the contents of a Multiwfn vtx.pdb file.

    Extracts the vertices and faces of the surface.

    Args:
        file: Name of file containing the vertices.

    Attributes:
        faces: Faces of surface
        vertices: Vertices of surface
    """

    faces: Array2DInt | None
    vertices: Array1DFloat

    def __init__(self, file: str | PathLike) -> None:  # noqa: C901
        # Parse file to see if it containts connectivity
        with open(file) as file:
            lines = file.readlines()

        # Get the number of vertices
        n_vertices = int(lines[0].strip().split()[5])

        # Parse the vertex positions and their connectivities
        vertices = {}
        vertex_map = {}
        connectivities: dict[int, set[int]] = {
            i: set() for i in range(1, n_vertices + 1)
        }
        vertex_counter = 1
        included_vertex_counter = 1
        for line in lines:
            if "HETATM" in line:
                if line[13] == "C":
                    x = float(line[32:39])
                    y = float(line[40:47])
                    z = float(line[48:55])
                    vertices[vertex_counter] = [x, y, z]
                    vertex_map[vertex_counter] = included_vertex_counter
                    included_vertex_counter += 1
                vertex_counter += 1
            if "CONECT" in line:
                n_entries = int(len(line.strip()) / 6 - 1)
                entries = []
                for i in range(1, n_entries + 1):
                    entry = int(line[i * 6 : i * 6 + 6])
                    entries.append(entry)
                connectivities[entries[0]].update(entries[1:])

        # Establish faces based on connectivity
        # https://stackoverflow.com/questions/1705824/finding-cycle-of-3-nodes-or-triangles-in-a-graph # noqa: B950
        if any(connectivities.values()):
            faces = []
            visited = set()
            for vertex_1 in connectivities:
                temp_visited = set()
                for vertex_2 in connectivities[vertex_1]:
                    if vertex_2 in visited:
                        continue
                    for vertex_3 in connectivities[vertex_2]:
                        if vertex_3 in visited or vertex_3 in temp_visited:
                            continue
                        if vertex_1 in connectivities[vertex_3]:
                            triangle_vertices = [vertex_1, vertex_2, vertex_3]
                            mapped_vertices = [
                                vertex_map[vertex] - 1 for vertex in triangle_vertices
                            ]
                            faces.append(mapped_vertices)
                    temp_visited.add(vertex_2)
                visited.add(vertex_1)
            self.faces = np.array(faces)
        else:
            self.faces = None

        # Set up attributes.
        self.vertices = np.array(list(vertices.values()))

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({len(self.vertices)!r} points)"


def read_gjf(file: str | PathLike) -> tuple[Array1DAny, Array1DFloat]:
    """Reads Gaussian gjf/com file and returns elements and coordinates.

    Args:
        file: gjf/com file object.

    Returns:
        elements: Elements as atomic symbols or numbers
        coordinates: Coordinates (Å)
    """
    # Read file and split lines
    with open(file) as f:
        text = f.read()
    xyz_lines = text.split("\n\n")[2].splitlines()[1:]
    xyz_lines = [line for line in xyz_lines if line.strip()]

    # Loop over lines and store elements and coordinates
    elements: list[int | str] = []
    coordinates: list[list[float]] = []
    for line in xyz_lines:
        split_line = line.strip().split()
        element = split_line[0]
        if element.isdigit():
            elements.append(int(element))
        else:
            elements.append(element)
        atom_coordinates = [float(i) for i in split_line[1:]]
        coordinates.append(atom_coordinates)
    elements: Array1DInt | Array1DStr = np.array(elements)
    coordinates: Array2DFloat = np.array(coordinates)

    return elements, coordinates


def read_geometry(
    file: str | PathLike,
) -> tuple[Array1DInt | Array1DStr, Array2DFloat | Array3DFloat]:
    """Read geometry file and guess parser based on suffix.

    Args:
        file: Filename or Path object

    Returns:
        elements: Elements as atomic symbols or numbers
        coordinates: Coordinates (Å)

    Raises:
        ValueError: When suffix is not supported
    """
    if isinstance(file, str):
        suffix = file.lower().split(".")[-1]
    elif isinstance(file, PathLike):
        path = Path(file)
        suffix = path.suffix[1:]

    if suffix == "xyz":
        elements, coordinates = read_xyz(file)
    elif suffix in ("gjf", "com"):
        elements, coordinates = read_gjf(file)
    else:
        raise ValueError(f"File suffix {suffix} not supported.")

    return elements, coordinates


def read_xyz(
    file: str | PathLike,
) -> tuple[Array1DInt | Array1DStr, Array2DFloat | Array3DFloat]:
    """Reads xyz file.

    Returns elements as written (atomic numbers or symbols) and coordinates.

    Args:
        file: Filename or Path object

    Returns:
        elements: Elements as atomic symbols or numbers
        coordinates: Coordinates (Å)
    """
    # Read file and split lines
    with open(file) as f:
        lines = f.readlines()

    # Loop over lines and store elements and coordinates
    elements: list[int | str] = []
    coordinates: list[list[float]] = []
    n_atoms = int(lines[0].strip())
    line_chunks = zip(*[iter(lines)] * (n_atoms + 2))
    for line_chunk in line_chunks:
        for line in line_chunk[2:]:
            strip_line = line.strip().split()
            atom = strip_line[0]
            if atom.isdigit():
                atom = int(atom)
            elements.append(atom)
            coordinates.append(
                [float(strip_line[1]), float(strip_line[2]), float(strip_line[3])]
            )
    elements = np.array(elements)[:n_atoms]
    coordinates: Array2DFloat | Array3DFloat = np.array(coordinates).reshape(
        -1, n_atoms, 3
    )
    if coordinates.shape[0] == 1:
        coordinates = coordinates[0]

    return elements, coordinates


@requires_dependency([Import("cclib")], globals())
def read_cclib(file: Any) -> tuple[Array1DFloat, Array2DFloat | Array3DFloat]:
    """Reads geometry file with cclib.

    Returns elements as atomic numbers and coordinates.

    Args:
        file: Input file to cclib

    Returns:
        elements: Elements as atomic symbols or numbers
        coordinates: Coordinates (Å)
    """
    data = cclib.io.ccread(file)
    elements = data.atomnos
    coordinates = data.atomcoords

    if coordinates.shape[0] == 1:
        coordinates = coordinates[0]

    return elements, coordinates


def get_xyz_string(
    symbols: Sequence[str], coordinates: Sequence[Sequence[float]], comment: str = ""
) -> str:
    """Return xyz string.

    Args:
        symbols: Atomic symbols
        coordinates: Atomic coordinates (Å)
        comment: Comment

    Returns:
        string: XYZ string
    """
    string = f"{len(symbols)}\n"
    string += f"{comment}\n"
    for s, c in zip(symbols, coordinates):
        string += f"{s:10s}{c[0]:10.5f}{c[1]:10.5f}{c[2]:10.5f}\n"

    return string


def write_xyz(
    file: str | PathLike,
    elements: Iterable[int] | Iterable[str],
    coordinates: ArrayLike2D | ArrayLike3D,
    comments: Sequence[str] | None = None,
) -> None:
    """Writes xyz file from elements and coordinates.

    Args:
        file: xyz file or path object
        elements: Elements as atomic symbols or numbers
        coordinates: Coordinates (Å)
        comments: Comments
    """
    # Convert elements to symbols
    symbols = convert_elements(elements, output="symbols")
    coordinates: Array2DFloat | Array3DFloat = np.array(coordinates).reshape(
        -1, len(symbols), 3
    )
    if comments is None:
        comments = [""] * len(coordinates)

    # Write the xyz file
    with open(file, "w") as f:
        for coord, comment in zip(coordinates, comments):
            xyz_string = get_xyz_string(symbols, coord, comment=comment)
            f.write(xyz_string)


def write_xtb_inp(
    file: str | PathLike,
    instruction_blocks: MutableMapping[str, list[str]],
) -> None:
    """Write input instructions to xtb input file.

    Args:
        file: path to the xtb input file to create
        instruction_blocks: xtb input instructions to write in the file
    """
    string = ""
    for flag, content in instruction_blocks.items():
        string += f"${flag}\n"
        for line in content:
            string += f"   {line}\n"
    string += "$end\n"
    with open(file, "w") as f:
        f.write(string)



"""xtb interface code."""

from __future__ import annotations

from collections.abc import Iterable
from contextlib import nullcontext
from dataclasses import dataclass
import functools
import json
import os
from pathlib import Path
import re
import shutil
import subprocess
from tempfile import TemporaryDirectory
from typing import Any, cast

import numpy as np

from morfeus import config
from morfeus.data import (
    AU_TO_DEBYE,
    DEBYE_TO_AU,
    EV_TO_HARTREE,
    HARTREE_TO_EV,
    HARTREE_TO_KCAL,
    HARTREE_TO_KJ,
)
from morfeus.io import read_geometry, write_xtb_inp, write_xyz
from morfeus.typing import Array1DFloat, Array2DFloat, ArrayLike2D
from morfeus.utils import convert_elements, requires_executable


class XTB:
    """Calculate electronic properties with the xtb program.

    Args:
        elements: Elements as atomic symbols or numbers
        coordinates: Coordinates (Å)
        method: Method to use in xtb. Currently works with:
            - 2: GFN2-xTB (default)
            - 1: GFN1-xTB
            - ptb: PTB
        charge: Molecular charge
        n_unpaired: Number of unpaired electrons
        solvent: Solvent. Uses the ALPB solvation model
        electronic_temperature: Electronic temperature (K)
        n_processes: Number of parallel processes in xtb.
            If not provided, runs on 1 thread.
        env_variables: Optional dictionary to set environment variables for xtb execution
        run_path: Folder path to run xTB calculation. If not provided, runs in a temporary folder
    """

    _charge: int
    _coordinates: Array2DFloat
    _electronic_temperature: int | None
    _elements: Array1DFloat
    _n_unpaired: int | None
    _results: XTBResults
    _solvent: str | None
    _method: int | str
    _n_processes: int | None
    _env_variables: dict[str, str] | None

    _xyz_input_file: str = "xtb.xyz"
    _xtb_input_file: str = "xtb.inp"

    def __init__(
        self,
        elements: Iterable[int] | Iterable[str],
        coordinates: ArrayLike2D,
        method: int | str = 2,
        charge: int = 0,
        n_unpaired: int | None = None,
        solvent: str | None = None,
        electronic_temperature: int | None = None,
        n_processes: int | None = None,
        env_variables: dict[str, str] | None = None,
        run_path: Path | str | None = None,
    ) -> None:
        # Converting elements to atomic numbers if the are symbols
        self._elements = np.array(convert_elements(elements, output="numbers"))

        # Store settings
        self._coordinates = np.array(coordinates)
        self._method = str(method).lower()
        self._charge = charge
        self._solvent = solvent
        self._n_unpaired = n_unpaired
        self._electronic_temperature = electronic_temperature
        self._n_processes = n_processes
        self._env_variables = env_variables

        self._run_path = Path(run_path) if run_path else None

        self._default_xtb_command = (
            f"xtb {XTB._xyz_input_file} --json --chrg {self._charge}"
        )
        if self._method in ["1", "2"]:
            self._default_xtb_command += f" --gfn {int(self._method)}"
        elif self._method == "ptb":
            self._default_xtb_command += " --ptb"
        else:
            raise ValueError(
                f"Method {self._method!r} not supported. Choose between: 2, 1, or ptb."
            )
        if self._solvent is not None:
            if self._method == "ptb":
                raise ValueError(
                    "Solvation is not available with PTB. "
                    "Remove solvent or use another xTB method."
                )
            self._default_xtb_command += f" --alpb {self._solvent}"
        if self._n_unpaired is not None:
            self._default_xtb_command += f" --uhf {self._n_unpaired}"
        if self._electronic_temperature is not None:
            self._default_xtb_command += f" --etemp {self._electronic_temperature}"

        self._results = XTBResults()
        self._corrected: bool | None = None

    def get_bond_order(self, i: int, j: int) -> float:
        """Return bond order between two atoms.

        Args:
            i: Index of atom 1 (1-indexed)
            j: Index of atom 2 (1-indexed)

        Returns:
            Bond order between atoms i and j

        Raises:
            ValueError: If no bond exists between the given atoms.
        """
        bonds_orders = self.get_bond_orders()
        if (i, j) in bonds_orders:
            bond_order = bonds_orders[(i, j)]
        elif (j, i) in bonds_orders:
            bond_order = bonds_orders[(j, i)]
        else:
            raise ValueError(f"No bond order calculated between atoms {i} and {j}.")

        return bond_order

    def get_bond_orders(self) -> dict[tuple[int, int], float]:
        """Return bond orders."""
        if self._results.bond_orders is None:
            self._run_xtb("sp")
            self._results.bond_orders = cast(
                list[tuple[int, int, float]], self._results.bond_orders
            )
        bond_orders = {(x[0], x[1]): round(x[2], 12) for x in self._results.bond_orders}

        return bond_orders

    def get_charges(self, model: str = "Mulliken") -> dict[int, float]:
        """Return atomic charges.

        Args:
            model: Charge model to use.
                - 'Mulliken'
                - 'CM5': only available with GFN1-xTB

        Raises:
            ValueError: If given charge model is not available with the chosen xtb method.

        Returns:
            Atomic charges (1-indexed)
        """
        if model == "CM5":
            if self._method != "1":
                raise ValueError("CM5 charge model is only available with GFN1-xTB.")
            elif self._results.charges_cm5 is None:
                self._run_xtb("sp")
                self._results.charges_cm5 = cast(list[float], self._results.charges_cm5)
            charges = self._results.charges_cm5

        elif model == "Mulliken":
            if self._results.charges is None:
                self._run_xtb("sp")
                self._results.charges = cast(list[float], self._results.charges)
            charges = self._results.charges

        else:
            raise ValueError(
                f"Charge model {model!r} not supported. Choose between 'Mulliken' or 'CM5'."
            )

        return {i: charge for i, charge in enumerate(charges, start=1)}

    def get_energy(self) -> float:
        """Return total energy (Eh)."""
        if self._results.total_energy is None:
            self._run_xtb("sp")
            self._results.total_energy = cast(float, self._results.total_energy)

        return self._results.total_energy

    def get_homo(self, unit: str = "Eh") -> float:
        """Return HOMO energy.

        Args:
            unit: 'Eh', 'eV', 'kcal/mol' or kJ/mol'

        Returns:
            HOMO energy (Eh, eV, kcal/mol or kJ/mol)

        Raises:
            ValueError: If given unit is not supported
        """
        if self._results.homo is None:
            self._run_xtb("sp")
            self._results.homo = cast(dict[str, float], self._results.homo)

        if unit == "Eh":
            return self._results.homo["Eh"]
        elif unit == "eV":
            return self._results.homo["eV"]
        elif unit == "kcal/mol":
            return round(self._results.homo["Eh"] * HARTREE_TO_KCAL, 7)
        elif unit == "kJ/mol":
            return round(self._results.homo["Eh"] * HARTREE_TO_KJ, 7)
        else:
            raise ValueError("Unit must be either 'Eh', 'eV', 'kcal/mol' or kJ/mol'.")

    def get_lumo(self, unit: str = "Eh") -> float:
        """Return LUMO energy.

        Args:
            unit: 'Eh', 'eV', 'kcal/mol' or kJ/mol'

        Returns:
            LUMO energy (Eh, eV, kcal/mol or kJ/mol)

        Raises:
            ValueError: If given unit is not supported
        """
        if self._results.lumo is None:
            self._run_xtb("sp")
            self._results.lumo = cast(dict[str, float], self._results.lumo)

        if unit == "Eh":
            return self._results.lumo["Eh"]
        elif unit == "eV":
            return self._results.lumo["eV"]
        elif unit == "kcal/mol":
            return round(self._results.lumo["Eh"] * HARTREE_TO_KCAL, 7)
        elif unit == "kJ/mol":
            return round(self._results.lumo["Eh"] * HARTREE_TO_KJ, 7)
        else:
            raise ValueError("Unit must be either 'Eh', 'eV', 'kcal/mol' or kJ/mol'.")

    def get_homo_lumo_gap(self, unit: str = "eV") -> float:
        """Return HOMO-LUMO gap.

        Args:
            unit: 'Eh', 'eV', 'kcal/mol' or kJ/mol'

        Returns:
            HOMO-LUMO gap (Eh, eV, kcal/mol or kJ/mol)

        Raises:
            ValueError: If given unit is not supported
        """
        if self._results.gap is None:
            self._run_xtb("sp")
            self._results.gap = cast(float, self._results.gap)

        if unit == "eV":
            gap = self._results.gap
        elif unit == "Eh":
            gap = round(self._results.gap * EV_TO_HARTREE, 8)
        elif unit == "kcal/mol":
            gap = round(self._results.gap * EV_TO_HARTREE * HARTREE_TO_KCAL, 8)
        elif unit == "kJ/mol":
            gap = round(self._results.gap * EV_TO_HARTREE * HARTREE_TO_KJ, 8)
        else:
            raise ValueError("Unit must be either 'Eh', 'eV', 'kcal/mol' or kJ/mol'.")

        return gap

    def get_fermi_level(self) -> float:
        """Return Fermi level (eV)."""
        if self._results.fermi_level is None:
            self._run_xtb("sp")
            self._results.fermi_level = cast(float, self._results.fermi_level)

        return self._results.fermi_level

    def get_dipole(self) -> Array1DFloat:
        """Return molecular dipole vector (a.u.)."""
        if self._results.dipole_vect is None:
            self._run_xtb("sp")

        return self._results.dipole_vect

    def get_dipole_moment(self, unit: str = "au") -> float:
        """Return molecular dipole moment.

        Args:
            unit: 'au' or 'debye'

        Returns:
            Molecular dipole moment (a.u. or debye)

        Raises:
            ValueError: If given unit is not supported
        """
        if self._results.dipole_moment is None:
            self._run_xtb("sp")
            self._results.dipole_moment = cast(float, self._results.dipole_moment)

        if unit == "debye":
            return self._results.dipole_moment
        elif unit == "au":
            return round(self._results.dipole_moment * DEBYE_TO_AU, 3)
        else:
            raise ValueError("Unit must be either 'au' or 'debye'.")

    def get_atom_dipoles(self) -> dict[int, Array1DFloat]:
        """Return atomic dipole vectors (a.u.).

        Raises:
            ValueError: If the chosen method is GFN1-xTB (does not support atomic dipoles)

        Returns:
            Atomic dipole vectors (a.u.)
        """
        if self._method == "1":
            raise ValueError(
                "Atomic dipoles are not available with GFN1-xTB. Choose another xtb method."
            )

        if self._results.atom_dipole_vect is None:
            self._run_xtb("sp")
            self._results.atom_dipole_vect = cast(
                np.ndarray, self._results.atom_dipole_vect
            )
        atom_dipole_vectors = {
            i: row for i, row in enumerate(self._results.atom_dipole_vect, start=1)
        }

        return atom_dipole_vectors

    def get_atom_dipole_moments(self, unit: str = "au") -> dict[int, float]:
        """Return atomic dipole moments.

        Args:
            unit: 'au' or 'debye'

        Returns:
            Atomic dipole moments (a.u. or debye)

        Raises:
            ValueError: If given unit is not supported
        """
        dipole_vectors = self.get_atom_dipoles()
        dipole_moments = {
            i: round(float(np.linalg.norm(vect)), 8)
            for i, vect in dipole_vectors.items()
        }
        if unit == "debye":
            dipole_moments = {
                i: round(norm * AU_TO_DEBYE, 8) for i, norm in dipole_moments.items()
            }
        elif unit != "au":
            raise ValueError("Unit must be either 'au' or 'debye'.")

        return dipole_moments

    def get_atom_polarizabilities(self) -> dict[int, float]:
        """Return atomic polarizabilities.

        Raises:
            ValueError: If the chosen method is not GFN2-xTB
                (necessary for polarizability calculations)

        Returns:
            Atomic polarizabilities
        """
        if self._method != "2":
            raise ValueError("Polarizability is only available with GFN2-xTB.")

        if self._results.atom_polarizabilities is None:
            self._run_xtb("sp")
            self._results.atom_polarizabilities = cast(
                list[float], self._results.atom_polarizabilities
            )
        atom_polarizabilities = {
            i: polar
            for i, polar in enumerate(self._results.atom_polarizabilities, start=1)
        }

        return atom_polarizabilities

    def get_molecular_polarizability(self) -> float:
        """Return molecular polarizability.

        Raises:
            ValueError: If the chosen method is not GFN2-xTB
                (necessary for polarizability calculations)

        Returns:
            Molecular polarizability
        """
        if self._method != "2":
            raise ValueError("Polarizability is only available with GFN2-xTB.")

        if self._results.mol_polarizability is None:
            self._run_xtb("sp")
            self._results.mol_polarizability = cast(
                float, self._results.mol_polarizability
            )

        return self._results.mol_polarizability

    def get_ip(self, corrected: bool = True) -> float:
        """Return ionization potential.

        Args:
            corrected: Whether to apply empirical correction term

        Returns:
            Ionization potential (eV)
        """
        if self._results.ip is None or self._corrected != corrected:
            self._corrected = corrected
            self._run_xtb("ipea")
            self._results.ip = cast(float, self._results.ip)

        return self._results.ip

    def get_ea(self, corrected: bool = True) -> float:
        """Return electron affinity.

        Args:
            corrected: Whether to apply empirical correction term

        Returns:
            Electron affinity (eV)
        """
        if self._results.ea is None or self._corrected != corrected:
            self._corrected = corrected
            self._run_xtb("ipea")
            self._results.ea = cast(float, self._results.ea)

        return self._results.ea

    def get_chemical_potential(self, corrected: bool = True) -> float:
        """Calculate chemical potential.

        Args:
            corrected: Whether to apply empirical correction term

        Returns:
            Chemical potential (eV)
        """
        if self._results.ip is None or self._corrected != corrected:
            self._corrected = corrected
            self._run_xtb("ipea")
        chem_pot = (
            -(self.get_ip(corrected=corrected) + self.get_ea(corrected=corrected)) / 2
        )

        return chem_pot

    def get_electronegativity(self, corrected: bool = True) -> float:
        """Calculate electronegativity.

        Args:
            corrected: Whether to apply empirical correction term

        Returns:
            Electronegativity (eV)
        """
        chem_pot = self.get_chemical_potential(corrected=corrected)

        return -chem_pot

    def get_hardness(self) -> float:
        """Calculate hardness (eV)."""
        hardness = round(self.get_ip() - self.get_ea(), 4)

        return hardness

    def get_softness(self) -> float:
        """Calculate softness (eV)."""
        hardness = self.get_hardness()

        return round(1 / hardness, 4)

    def get_fukui(self, variety: str, corrected: bool = True) -> dict[int, float]:
        """Calculate Fukui coefficients.

        Args:
            variety: Type of Fukui coefficient:'nucleophilicity' = 'minus',
                'electrophilicity' = 'plus', 'radical', 'dual',
                'local_nucleophilicity' or 'local_electrophilicity'.
                Note: 'nucleophilicity' and 'electrophilicity' are synonym for respectively
                'minus' and 'plus'.
            corrected: Whether to apply empirical correction term to the inonization potential
                and electron affinity (only applicable for local electrophilicity calculation)

        Returns:
            fukui: Atomic Fukui coefficients

        Raises:
            ValueError: When variety does not exist
        """
        if self._results.fukui_plus is None:
            self._run_xtb("fukui")

        varieties = [
            "minus",
            "nucleophilicity",
            "plus",
            "electrophilicity",
            "radical",
            "dual",
            "local_electrophilicity",
            "local_nucleophilicity",
        ]
        fukui: Array1DFloat
        if variety in ["local_nucleophilicity", "nucleophilicity", "minus"]:
            fukui = self._results.fukui_minus
        elif variety in ["electrophilicity", "plus"]:
            fukui = self._results.fukui_plus
        elif variety == "radical":
            fukui = self._results.fukui_radical
        elif variety == "dual":
            fukui = np.around(
                np.array(self._results.fukui_plus)
                - np.array(self._results.fukui_minus),
                3,
            )
        elif variety == "local_electrophilicity":
            fukui_radical: Array1DFloat = np.array(self._results.fukui_radical)
            fukui_dual: Array1DFloat = np.array(self._results.fukui_plus) - np.array(
                self._results.fukui_minus
            )
            chem_pot = self.get_chemical_potential(corrected=corrected)
            hardness = self.get_hardness()
            fukui = (
                -(chem_pot / hardness) * fukui_radical
                + 1 / 2 * (chem_pot / hardness) ** 2 * fukui_dual
            )
        else:
            raise ValueError(
                f"Variety {variety!r} does not exist. "
                f"Choose one of {', '.join(varieties)}. "
                "Note: 'nucleophilicity' and 'electrophilicity' are synonym for "
                "respectively 'minus' and 'plus'."
            )

        fukui = {i: float(fukui) for i, fukui in enumerate(fukui, start=1)}

        return fukui

    def get_global_descriptor(self, variety: str, corrected: bool = True) -> float:
        """Calculate global reactivity descriptors.

        Args:
            corrected: Whether to apply empirical correction term
            variety: Type of descriptor: 'electrophilicity', 'nucleophilicity',
                'electrofugality' or 'nucleofugality'

        Returns:
            descriptor: Global reactivity descriptor (eV)

        Raises:
            ValueError: When variety does not exist
        """
        varieties = [
            "electrofugality",
            "electrophilicity",
            "nucleofugality",
            "nucleophilicity",
        ]
        if variety == "electrophilicity":
            descriptor = (
                self.get_ip(corrected=corrected) + self.get_ea(corrected=corrected)
            ) ** 2 / (
                8
                * (self.get_ip(corrected=corrected) - self.get_ea(corrected=corrected))
            )
        elif variety == "nucleophilicity":
            descriptor = -self.get_ip(corrected=corrected)
        elif variety == "electrofugality":
            descriptor = (
                3 * self.get_ip(corrected=corrected) - self.get_ea(corrected=corrected)
            ) ** 2 / (
                8
                * (self.get_ip(corrected=corrected) - self.get_ea(corrected=corrected))
            )
        elif variety == "nucleofugality":
            descriptor = (
                self.get_ip(corrected=corrected) - 3 * self.get_ea(corrected=corrected)
            ) ** 2 / (
                8
                * (self.get_ip(corrected=corrected) - self.get_ea(corrected=corrected))
            )
        else:
            raise ValueError(
                f"Variety {variety!r} does not exist. "
                f"Choose one of {', '.join(varieties)}."
            )

        return descriptor

    def get_s_pop(self) -> dict[int, float]:
        """Return atomic population partitioned to the s shell from Mulliken population analysis.

        Raises:
            ValueError: If the chosen method is not GFN1-xTB
                (necessary for population calculations)

        Returns:
            Atomic s shell populations
        """
        if self._method != "1":
            raise ValueError("Shell populations are only available with GFN1-xTB.")
        elif self._results.s_pop is None:
            self._run_xtb("sp")
            self._results.s_pop = cast(list[float], self._results.s_pop)
        s_pop = self._results.s_pop

        return {i: pop for i, pop in enumerate(s_pop, start=1)}

    def get_p_pop(self) -> dict[int, float]:
        """Return atomic population partitioned to the p shell from Mulliken population analysis.

        Raises:
            ValueError: If the chosen method is not GFN1-xTB
                (necessary for population calculations)

        Returns:
            Atomic p shell populations
        """
        if self._method != "1":
            raise ValueError("Shell populations are only available with GFN1-xTB.")
        elif self._results.p_pop is None:
            self._run_xtb("sp")
            self._results.p_pop = cast(list[float], self._results.p_pop)
        p_pop = self._results.p_pop

        return {i: pop for i, pop in enumerate(p_pop, start=1)}

    def get_d_pop(self) -> dict[int, float]:
        """Return atomic population partitioned to the d shell from Mulliken population analysis.

        Raises:
            ValueError: If the chosen method is not GFN1-xTB
                (necessary for population calculations)

        Returns:
            Atomic d shell populations
        """
        if self._method != "1":
            raise ValueError("Shell populations are only available with GFN1-xTB.")
        elif self._results.d_pop is None:
            self._run_xtb("sp")
            self._results.d_pop = cast(list[float], self._results.d_pop)
        d_pop = self._results.d_pop

        return {i: pop for i, pop in enumerate(d_pop, start=1)}

    def get_covcn(self) -> dict[int, float]:
        """Return atomic covalent coordination numbers.

        Raises:
            ValueError: If the chosen method is not GFN2-xTB (necessary for covCN calculations)

        Returns:
            Atomic covalent coordination numbers
        """
        if self._method != "2":
            raise ValueError(
                "Covalent coordination number is only available with GFN2-xTB."
            )

        if self._results.covcn is None:
            self._run_xtb("sp")
            self._results.covcn = cast(list[float], self._results.covcn)
        covcns = self._results.covcn

        return {i: covcn for i, covcn in enumerate(covcns, start=1)}

    def get_fod_population(self) -> dict[int, float]:
        """Return atomic fractional occupation number weighted density population.

        The FOD calculation is performed by default with an electronic temperature of 5000 K.

        Returns:
            Atomic FOD populations
        """
        if self._results.fod_pop is None:
            self._run_xtb("fod")
            self._results.fod_pop = cast(list[float], self._results.fod_pop)

        fod_pop = {i: pop for i, pop in enumerate(self._results.fod_pop, start=1)}

        return fod_pop

    def get_nfod(self) -> float:
        """Return NFOD descriptor.

        NFOD is the integration over all space of the fractional occupation number
        weighted density (FOD).
        The FOD calculation is performed by default with an electronic temperature of 5000 K.

        Returns:
            NFOD descriptor
        """
        fod_pop = self.get_fod_population()
        nfod = sum(fod_pop.values())

        return nfod

    def get_solvation_energy(self, unit: str = "Eh") -> float:
        """Return solvation free energy.

        Args:
            unit: 'Eh', 'eV', 'kcal/mol' or kJ/mol'

        Returns:
            Solvation free energy (Eh, eV, kcal/mol or kJ/mol)

        Raises:
            ValueError: If no solvent is specified (necessary for solvation calculations)
            ValueError: If given unit is not supported
        """
        if self._solvent is None:
            raise ValueError("Solvation energy is only available with solvent.")

        if self._results.g_solv is None:
            self._run_xtb("sp")
            self._results.g_solv = cast(float, self._results.g_solv)

        if unit == "Eh":
            return self._results.g_solv
        elif unit == "kcal/mol":
            return round(self._results.g_solv * HARTREE_TO_KCAL, 12)
        elif unit == "kJ/mol":
            return round(self._results.g_solv * HARTREE_TO_KJ, 12)
        elif unit == "eV":
            return round(self._results.g_solv * HARTREE_TO_EV, 12)
        else:
            raise ValueError("Unit must be either 'Eh', 'eV', 'kcal/mol' or kJ/mol'.")

    def get_solvation_h_bond_correction(self, unit: str = "Eh") -> float:
        """Return hydrogen bonding correction to the solvation free energy.

        The hydrogen bonding correction is 0.0 for non-polar solvents.

        Args:
            unit: 'Eh', 'eV', 'kcal/mol' or kJ/mol'

        Returns:
            Hydrogen bonding correction to the solvation free energy (Eh, eV, kcal/mol or kJ/mol)

        Raises:
            ValueError: If no solvent is specified (necessary for solvation calculations)
            ValueError: If given unit is not supported
        """
        if self._solvent is None:
            raise ValueError(
                "Hydrogen bonding correction to solvation is only available with solvent."
            )

        if self._results.g_solv_hb is None:
            self._run_xtb("sp")
            self._results.g_solv_hb = cast(float, self._results.g_solv_hb)

        if unit == "Eh":
            return self._results.g_solv_hb
        elif unit == "kcal/mol":
            return round(self._results.g_solv_hb * HARTREE_TO_KCAL, 12)
        elif unit == "kJ/mol":
            return round(self._results.g_solv_hb * HARTREE_TO_KJ, 12)
        elif unit == "eV":
            return round(self._results.g_solv_hb * HARTREE_TO_EV, 12)
        else:
            raise ValueError("Unit must be either 'Eh', 'eV', 'kcal/mol' or kJ/mol'.")

    def get_atomic_h_bond_corrections(self, unit: str = "Eh") -> dict[int, float]:
        """Calculate atomic hydrogen bonding corrections to the solvation free energy.

        Caveat: Due to the limited print precision of the H-bond terms outputted by xtb,
        the atomic energy corrections returned here have an error
        of max ± (atomic charge)^2 * 0.0005 Eh

        Args:
            unit: 'Eh', 'eV', 'kcal/mol' or kJ/mol'

        Returns:
            Atomic hydrogen bonding corrections to the solvation free energy
            (Eh, eV, kcal/mol or kJ/mol)

        Raises:
            ValueError: If no solvent is specified (necessary for solvation calculations)
            ValueError: If the specified solvent is not polar
                (necessary for atomic hydrogen bonding contributions)
            ValueError: If given unit is not supported
        """
        if self._solvent is None:
            raise ValueError(
                "Atomic hydrogen bonding corrections are only available with (polar) solvent."
            )

        if self._results.atom_hb_terms is None:
            self._run_xtb("sp")
            self._results.atom_hb_terms = cast(list[float], self._results.atom_hb_terms)

        if self._results.atom_hb_terms == []:
            raise ValueError(
                f"No hydrogen bonding contributions calculated with "
                f"{self._solvent!r}. Provide a polar solvent."
            )

        if self._method == "2":
            charges = self._results.charges
        elif self._method == "1":
            charges = self._results.charges_cm5
        charges = cast(list[float], charges)

        if unit == "Eh":
            conversion = 1.0
        elif unit == "kcal/mol":
            conversion = HARTREE_TO_KCAL
        elif unit == "kJ/mol":
            conversion = HARTREE_TO_KJ
        elif unit == "eV":
            conversion = HARTREE_TO_EV
        else:
            raise ValueError("Unit must be either 'Eh', 'eV', 'kcal/mol' or kJ/mol'.")

        h_bond_corrections = {
            i: round(hb_term * charge**2 * conversion, 8)
            for i, (hb_term, charge) in enumerate(
                zip(self._results.atom_hb_terms, charges), start=1
            )
        }

        return h_bond_corrections

    @requires_executable(["xtb"])
    def _run_xtb(self, runtype: str) -> None:  # noqa: C901
        """Run xTB calculation and parse results.

        Args:
            runtype: Type of calculation to perform:
                - 'sp': Single point calculation
                - 'ipea': Ionization potential and electron affinity calculation
                - 'fukui': Fukui coefficient calculation
                - 'fod': Fractional occupation density calculation

        Raises:
            ValueError: If runtype does not exist
            ValueError: If the xtb method chosen is not compatible with the requested calculation
            RuntimeError: If the xtb calculation failed
        """
        # Set xtb command
        runtypes = ["sp", "ipea", "fukui", "fod"]
        if runtype == "sp":
            command = self._default_xtb_command
            if self._solvent is not None:
                command += f" --input {XTB._xtb_input_file}"
        elif self._method == "ptb":
            raise ValueError(
                "PTB can only be used for calculations of bond orders, charges, dipole, "
                "and HOMO/LUMO energies.\n"
                "For other descriptors, choose another xtb method."
            )
        elif runtype == "ipea":
            command = self._default_xtb_command + " --vipea"
        elif runtype == "fukui":
            command = self._default_xtb_command + " --vfukui"
        elif runtype == "fod":
            command = self._default_xtb_command + " --fod"
        else:
            raise ValueError(
                f"Runtype {runtype!r} does not exist. Choose one of {', '.join(runtypes)}."
            )

        # Create temporary directory or use provided path
        tmp_dir_context = (
            TemporaryDirectory() if self._run_path is None else nullcontext()
        )
        with tmp_dir_context as tmp_dir:
            run_folder = (
                Path(cast(str, tmp_dir))
                if self._run_path is None
                else self._run_path / runtype
            )
            if self._run_path:
                if run_folder.exists():
                    shutil.rmtree(run_folder)
                run_folder.mkdir(parents=True)

            # Write xyz input file
            xyz_file = run_folder / XTB._xyz_input_file
            write_xyz(xyz_file, self._elements, self._coordinates)

            # To
            if self._solvent is not None:
                xtb_inp = run_folder / XTB._xtb_input_file
                write_xtb_inp(xtb_inp, {"write": ["gbsa=true"]})

            # Run xtb
            with open(run_folder / "xtb.out", "w") as stdout, open(
                run_folder / "xtb.err", "w"
            ) as stderr:
                env = self._set_env()
                subprocess.run(
                    command.split(),
                    cwd=run_folder,
                    stdout=stdout,
                    stderr=stderr,
                    env=env,
                )

            # Return error if xtb fails
            with open(run_folder / "xtb.err", "r", encoding="utf8") as f:
                err_content = f.read()
            if not re.search(r"(?<!ab)normal termination of xtb", err_content):
                with open(run_folder / "xtb.out", "r", encoding="utf8") as f:
                    out_content = f.read()
                    start_error_idx = out_content.find("######")
                    error = (
                        out_content[start_error_idx:]
                        if start_error_idx != -1
                        else err_content
                    )
                raise RuntimeError(f"xtb calculation failed. Error:\n{error}")

            # Extract results from xtb files
            if runtype == "sp":
                self._parse_json(run_folder / "xtbout.json")
                self._parse_wbo(run_folder / "wbo")
                self._parse_out_sp(run_folder / "xtb.out")
            elif runtype == "ipea":
                self._parse_out_ipea(run_folder / "xtb.out")
            elif runtype == "fukui":
                self._parse_out_fukui(run_folder / "xtb.out")
            elif runtype == "fod":
                self._parse_fod(run_folder / "fod")

    def _set_env(self) -> dict[str, str]:
        """Set environment variables for xTB execution."""
        if self._env_variables is not None:
            env = self._env_variables
        else:
            env = dict(os.environ)
            num_threads = (
                self._n_processes
                if self._n_processes is not None
                else config.OMP_NUM_THREADS
            )
            env["OMP_NUM_THREADS"] = f"{num_threads},1"
            env["MKL_NUM_THREADS"] = f"{num_threads}"
            env["OMP_STACKSIZE"] = config.OMP_STACKSIZE
            env["OMP_MAX_ACTIVE_LEVELS"] = str(config.OMP_MAX_ACTIVE_LEVELS)

        return env

    def _parse_json(self, json_file: Path | str) -> None:
        """Parse 'xtbout.json' file."""
        with open(json_file, "r", encoding="utf8") as f:

            # The code below fixes an error in the json file outputed
            # when running PTB with xtb 6.7.1
            # TODO: Remove/update when new xtb version is released
            lines_without_error = [line for line in f if line.strip() != ","]
            json_fixed = "".join(lines_without_error)
            data = json.loads(json_fixed)

        self._results.charges = data["partial charges"]
        self._results.gap = data["HOMO-LUMO gap / eV"]
        self._results.total_energy = data["total energy"]
        self._results.dipole_vect = np.array(data["dipole / a.u."])
        if self._method != "1":
            self._results.atom_dipole_vect = np.array(data["atomic dipole moments"])

    def _parse_wbo(self, wbo_file: Path | str) -> None:
        """Parse 'wbo' file."""
        wbos = []
        with open(wbo_file, "r", encoding="utf8") as f:
            for line in f:
                columns = line.split()
                wbos.append((int(columns[0]), int(columns[1]), float(columns[2])))
        self._results.bond_orders = wbos

    def _parse_out_sp(self, out_file: Path | str) -> None:  # noqa: C901
        """Parse 'xtb.out' file from xtb sp calculation."""
        with open(out_file, "r", encoding="utf8") as f:
            lines = f.readlines()
        (
            homo,
            lumo,
            atom_polarizabilities,
            covcn,
            atom_hb_terms,
            cm5_charges,
            s_pop,
            p_pop,
            d_pop,
        ) = (
            {},
            {},
            [],
            [],
            [],
            [],
            [],
            [],
            [],
        )
        in_polarizability_block, in_gbsa_block, in_cm5_block = False, False, False
        for i, line in enumerate(lines):
            if "(HOMO)" in line:
                homo["Eh"] = float(line.split()[-3])
                homo["eV"] = float(line.split()[-2])
            elif "(LUMO)" in line:
                lumo["Eh"] = float(line.split()[-3])
                lumo["eV"] = float(line.split()[-2])
            elif "Fermi-level" in line:
                fermi_level = float(line.split()[-2])
            elif "dipole" in line:
                if self._method == "2":
                    dipole_line = lines[i + 3].split()
                    dipole_moment = float(dipole_line[-1])
                elif self._method == "1":
                    dipole_line = lines[i + 2].split()
                    dipole_moment = float(dipole_line[-1])
                elif self._method == "ptb":
                    if "Total dipole" in line:
                        dipole_line = lines[i + 1].split()
                        dipole_moment = float(dipole_line[-1])
            elif self._method == "2" and "α(0)" in line:
                if "Mol." in line:
                    mol_polarizability = float(line.split()[-1])
                else:
                    in_polarizability_block = True
            elif in_polarizability_block:
                if line.strip():
                    atom_polarizabilities.append(float(line.split()[-1]))
                    covcn.append(float(line.split()[3]))
                else:
                    in_polarizability_block = False

            elif "Gsolv" in line:
                g_solv = float(line.split()[-3])
            elif "Ghb" in line:
                g_solv_hb = float(line.split()[-3])
            elif "H-bond" in line and "correction" not in line:
                in_gbsa_block = True
            elif in_gbsa_block:
                if line.strip():
                    atom_hb_terms.append(float(line.split()[-1]))
                else:
                    in_gbsa_block = False

            elif "Mulliken/CM5 charges" in line:
                in_cm5_block = True
            elif in_cm5_block:
                if line.strip():
                    cm5_charges.append(float(line.split()[2]))
                    s_pop.append(float(line.split()[-3]))
                    p_pop.append(float(line.split()[-2]))
                    d_pop.append(float(line.split()[-1]))
                else:
                    in_cm5_block = False

        self._results.homo = homo
        self._results.lumo = lumo
        self._results.fermi_level = fermi_level
        self._results.dipole_moment = dipole_moment
        if self._method == "2":
            self._results.atom_polarizabilities = atom_polarizabilities
            self._results.covcn = covcn
            self._results.mol_polarizability = mol_polarizability
        if self._solvent is not None:
            self._results.g_solv = g_solv
            self._results.g_solv_hb = g_solv_hb
            self._results.atom_hb_terms = atom_hb_terms
        if self._method == "1":
            self._results.charges_cm5 = cm5_charges
            self._results.s_pop = s_pop
            self._results.p_pop = p_pop
            self._results.d_pop = d_pop

    def _parse_out_ipea(self, out_file: Path | str) -> None:
        """Parse 'xtb.out' file from xtb ipea calculation."""
        with open(out_file, "r", encoding="utf8") as f:
            ip, ea, shift = None, None, None
            for line in f:
                if "delta SCC IP (eV)" in line:
                    ip = float(line.split()[-1])
                elif "delta SCC EA (eV)" in line:
                    ea = float(line.split()[-1])
                elif "empirical IP shift (eV)" in line:
                    shift = float(line.split()[-1])
                if ip and ea and shift:
                    break

        if ip is None or ea is None or shift is None:
            raise ValueError(f"Failed to parse IP and/or EA from {out_file}.")

        if not self._corrected:
            ip += shift
            ea += shift

        self._results.ip = ip
        self._results.ea = ea

    def _parse_out_fukui(self, out_file: Path | str) -> None:
        """Parse 'xtb.out' file from xtb fukui calculation."""
        with open(out_file, "r", encoding="utf8") as f:
            fukui_plus, fukui_minus, fukui_radical = [], [], []
            in_fukui_block = False
            for line in f:
                if "Fukui functions:" in line:
                    in_fukui_block = True
                if in_fukui_block:
                    if "-------------" in line or not line.strip():
                        in_fukui_block = False
                        break
                    if "Fukui functions" not in line and "#" not in line:
                        columns = line.split()
                        fukui_plus.append(float(columns[-3]))
                        fukui_minus.append(float(columns[-2]))
                        fukui_radical.append(float(columns[-1]))

        if not fukui_plus or not fukui_minus or not fukui_radical:
            raise ValueError(f"Failed to parse Fukui coefficients from {out_file}.")

        self._results.fukui_plus = fukui_plus
        self._results.fukui_minus = fukui_minus
        self._results.fukui_radical = fukui_radical

    def _parse_fod(self, fod_file: Path | str) -> None:
        """Parse 'fod' file."""
        with open(fod_file, "r", encoding="utf8") as f:
            lines = f.readlines()
        fod = [float(line.strip()) for line in lines]
        self._results.fod_pop = fod


@dataclass
class XTBResults:
    """Stores xTB descriptors."""

    charges: list[float] | None = None
    charges_cm5: list[float] | None = None
    bond_orders: list[tuple[int, int, float]] | None = None
    homo: dict[str, float] | None = None  # Stores both Eh and eV units
    lumo: dict[str, float] | None = None  # Stores both Eh and eV units
    gap: float | None = None  # Unit in eV
    atom_dipole_vect: Array2DFloat | None = None  # Unit in a.u.
    dipole_vect: Array1DFloat | None = None  # Unit in a.u.
    dipole_moment: float | None = None  # Unit in debye
    atom_polarizabilities: list[float] | None = None
    mol_polarizability: float | None = None
    total_energy: float | None = None  # Unit in Eh
    fermi_level: float | None = None  # Unit in eV
    s_pop: list[float] | None = None
    p_pop: list[float] | None = None
    d_pop: list[float] | None = None
    covcn: list[float] | None = None
    g_solv: float | None = None  # Unit in Eh
    g_solv_hb: float | None = None  # Unit in Eh
    atom_hb_terms: list[float] | None = None
    fod_pop: list[float] | None = None
    ip: float | None = None
    ea: float | None = None
    fukui_plus: list[float] | None = None
    fukui_minus: list[float] | None = None
    fukui_radical: list[float] | None = None


def cli(file: str) -> Any:
    """CLI for XTB.

    Args:
        file: Geometry file

    Returns:
        Partially instantiated class
    """
    elements, coordinates = read_geometry(file)
    return functools.partial(XTB, elements, coordinates)



"""Typing code for arrays.

Borrowed from gpaw/typing.py
"""

from __future__ import annotations

from typing import Any

from numpy import integer, signedinteger
from numpy.typing import ArrayLike, NDArray

IntLike = int | integer | signedinteger

ArrayLike1D = ArrayLike
ArrayLike2D = ArrayLike
ArrayLike3D = ArrayLike

ArrayND = NDArray
Array1DAny = Any
# TODO: Remove these when https://github.com/python/mypy/issues/11347 is resolved
Array1DFloat = Any  # NDArray[np.float_]
Array2DFloat = Any  # NDArray[np.float_]
Array3DFloat = Any  # NDArray[np.float_]
Array1DInt = Any  # NDArray[np.int_]
Array2DInt = Any  # NDArray[np.int_]
Array3DInt = Any  # NDArray[np.int_]
Array1DBool = Any  # NDArray[np.bool_]
Array1DStr = Any  # NDArray[np.str_]



"""Test Sterimol code."""

import csv
from pathlib import Path

from numpy.testing import assert_almost_equal
import pytest

from morfeus import read_gjf, Sterimol
from morfeus.utils import get_radii

DATA_DIR = Path(__file__).parent / "data" / "sterimol"


def test_H():
    """Test H substituent."""
    elements, coordinates = read_gjf(DATA_DIR / "gjfs" / "H.gjf")
    radii = get_radii(elements, radii_type="bondi")
    radii = [1.09 if radius == 1.20 else radius for radius in radii]
    sterimol = Sterimol(elements, coordinates, 1, 2, radii=radii)
    assert_almost_equal(
        (sterimol.L_value, sterimol.B_1_value, sterimol.B_5_value),
        (2.15, 1.09, 1.09),
        decimal=2,
    )


def pytest_generate_tests(metafunc):
    """Generate test data from csv file."""
    if "sterimol_data" in metafunc.fixturenames:
        with open(DATA_DIR / "reference_data.csv") as f:
            reader = csv.DictReader(f)
            records = list(reader)
        metafunc.parametrize("sterimol_data", records)


@pytest.mark.benchmark
def test_reference(sterimol_data):
    """Test against Sterimol reference data."""
    data = sterimol_data
    name = data["name"]
    L, B_1, B_5 = float(data["L"]), float(data["B_1"]), float(data["B_5"])

    # Use Paton's Bondi radii of 1.09 for H
    elements, coordinates = read_gjf(DATA_DIR / "gjfs" / f"{name}.gjf")
    radii = get_radii(elements, radii_type="bondi")
    radii = [1.09 if radius == 1.20 else radius for radius in radii]
    sterimol = Sterimol(elements, coordinates, 1, 2, radii=radii)

    assert_almost_equal(L, sterimol.L_value, decimal=2)
    assert_almost_equal(B_1, sterimol.B_1_value, decimal=2)
    assert_almost_equal(B_5, sterimol.B_5_value, decimal=2)



"""Conformer tools."""

from __future__ import annotations

from collections import Counter
from collections.abc import Iterable, Mapping, Sequence
from copy import copy, deepcopy
import functools
import numbers
from os import PathLike
from pathlib import Path
import subprocess
import tempfile
import typing
from typing import Any, cast, Type, TypeVar
import warnings

import numpy as np
from packaging.version import parse

from morfeus.data import (
    HARTREE,
    HARTREE_TO_KCAL,
    HARTREE_TO_KJ,
    K_B,
    KCAL_TO_HARTREE,
    KJ_TO_HARTREE,
)
from morfeus.io import CrestParser, write_xyz
from morfeus.qc import optimize_qc_engine, sp_qc_engine
from morfeus.typing import (
    Array1DFloat,
    Array1DInt,
    Array2DFloat,
    Array2DInt,
    Array3DFloat,
    ArrayLike1D,
    ArrayLike2D,
    ArrayLike3D,
)
from morfeus.utils import (
    convert_elements,
    Import,
    requires_dependency,
    requires_executable,
)

if typing.TYPE_CHECKING:
    import openbabel
    import openbabel.openbabel as ob
    import openbabel.pybel as pybel
    import rdkit
    from rdkit import Chem
    from rdkit.Chem import AllChem, Descriptors, rdDistGeom
    import spyrmsd


def boltzmann_average_dT(
    properties: ArrayLike1D, energies: ArrayLike1D, temperature: float = 298.15
) -> float:
    """Return the derivative of the Boltzmann average.

    Args:
        properties: Conformer properties
        energies: Conformer energies (a.u.)
        temperature: Temperature (K)

    Returns:
        derivative: Derivative of Boltzmann average.
    """
    energies: Array1DFloat = np.array(energies)
    properties: Array1DFloat = np.array(properties)

    # Calculate Boltzmann averaged properties
    avg_prop_en = boltzmann_statistic(
        properties * energies * HARTREE,
        energies,
        temperature=temperature,
        statistic="avg",
    )
    avg_en = boltzmann_statistic(
        energies * HARTREE, energies, temperature=temperature, statistic="avg"
    )
    avg_prop = boltzmann_statistic(
        properties, energies, temperature=temperature, statistic="avg"
    )

    # Calculate derivative
    derivative = (avg_prop_en - avg_en * avg_prop) / (K_B * temperature**2)

    return derivative


def boltzmann_statistic(
    properties: ArrayLike1D,
    energies: ArrayLike1D,
    temperature: float = 298.15,
    statistic: str = "avg",
) -> float:
    """Compute Boltzmann statistic.

    Args:
        properties: Conformer properties
        energies: Conformer energies (a.u.)
        temperature: Temperature (K)
        statistic: Statistic to compute: 'avg', 'var' or 'std'

    Returns:
        result: Boltzmann statistic

    Raises:
        ValueError: When statistic not specified correctly
    """
    properties: Array1DFloat = np.array(properties)

    # Get conformer weights
    weights = boltzmann_weights(energies, temperature)

    # Compute Boltzmann weighted statistic
    result: float
    if statistic == "avg":
        result = np.average(properties, weights=weights, axis=0)
    elif statistic == "var":
        avg = boltzmann_statistic(
            properties, energies, temperature=temperature, statistic="avg"
        )
        result = np.average((properties - avg) ** 2, weights=weights, axis=0)
    elif statistic == "std":
        avg = boltzmann_statistic(
            properties, energies, temperature=temperature, statistic="avg"
        )
        var = boltzmann_statistic(
            properties, energies, temperature=temperature, statistic="var"
        )
        result = np.sqrt(var)
    else:
        raise ValueError("Choose between: 'avg', 'var' and 'std'")

    return result


def boltzmann_weights(
    energies: ArrayLike1D, temperature: float = 298.15
) -> Array1DFloat:
    """Compute Boltzmann weights.

    Args:
        energies: Conformer energies (a.u.)
        temperature: Temperature (K)

    Returns:
        weights: Conformer weights (normalized to unity)
    """
    energies: Array1DFloat = np.array(energies)
    energies -= energies.min()
    terms = np.exp(-energies / (K_B / HARTREE * temperature))
    weights: Array1DFloat = terms / np.sum(terms)

    return weights


class Conformer:
    """Conformer with coordinates, energy and properties.

    Args:
        elements: Elements as atomic symbols or numbers
        coordinates: Coordinates (Å)
        energy: Energy (a.u.)
        degeneracy: Degeneracy
        properties: Conformers properties.
        cip_label: Tuple of CIP labels for all atoms: 'R', 'S' or ''

    Attributes:
        cip_label: Tuple of CIP labels for all atoms: 'R', 'S' or ''
        coordinates: Coordinates (Å)
        degeneracy: Degeneracy
        elements: Elements as atomic symbols or numbers
        energy: Energy (a.u.)
        properties: Conformers properties.
    """

    cip_label: tuple[str, ...] | None
    coordinates: Array2DFloat
    degeneracy: int
    elements: Array1DFloat
    energy: float | None
    properties: dict[str, float]

    def __init__(
        self,
        elements: Iterable[int] | Iterable[str],
        coordinates: ArrayLike2D,
        energy: float | None = None,
        degeneracy: int = 1,
        properties: dict[str, float] | None = None,
        cip_label: tuple[str] | None = None,
    ) -> None:
        if properties is None:
            properties = {}
        self.coordinates = np.array(coordinates)
        self.elements = np.array(elements)
        self.energy = energy
        self.properties = properties
        self.degeneracy = degeneracy
        self.cip_label = cip_label

    def __repr__(self) -> str:
        n_atoms = self.coordinates.shape[0]
        return f"{self.__class__.__name__}({n_atoms!r} atoms)"


T = TypeVar("T", bound="ConformerEnsemble")


class ConformerEnsemble:
    """Conformer ensemble object.

    Supports sorting, pruning, optimization and single-point calculations.

    Args:
        elements: Elements as atomic symbols or numbers
        conformer_coordinates: Conformer coordinates (Å)
        energies: Energy (a.u.)
        connectivity_matrix: Connectivity matrix
        degeneracies: Degeneracies
        properties: Conformers properties.
        charge: Molecular charge
        multiplicity: Molecular multiplicity.
        formal_charges: Atomic formal charges
        ref_cip_label: Tuple of CIP labels for all atoms: 'R', 'S' or ''

    Attributes:
        charge: Molecular charge
        conformers: Conformers
        connectivity_matrix: Connectivity matrix
        elements: Elements as atomic symbols or numbers
        formal_charges: Atomic formal charges.
        mol: RDKit mol object.
        multiplicity: Molecular multiplicity.
        ref_cip_label: Tuple of CIP labels for all atoms: 'R', 'S' or ''
    """

    charge: int
    confomers: list[Conformer]
    connectivity_matrix: Array2DInt
    elements: Array1DInt
    formal_charges: Array1DInt
    mol: Chem.Mol
    multiplicity: int
    ref_cip_label: tuple[str, ...] | None

    def __init__(
        self,
        elements: Iterable[int] | Iterable[str],
        conformer_coordinates: ArrayLike3D | None = None,
        energies: ArrayLike1D | None = None,
        connectivity_matrix: ArrayLike2D | None = None,
        degeneracies: ArrayLike1D | None = None,
        properties: Mapping[str, ArrayLike1D] | None = None,
        charge: int | None = None,
        multiplicity: int | None = None,
        formal_charges: ArrayLike1D | None = None,
        ref_cip_label: tuple[str, ...] | None = None,
    ) -> None:
        elements: Array1DInt = np.array(convert_elements(elements, output="numbers"))
        self.elements = elements

        if conformer_coordinates is None:
            conformer_coordinates = []
        if formal_charges is None:
            formal_charges = np.zeros(len(elements), dtype=int)
        if charge is None:
            charge = int(np.sum(formal_charges))
        else:
            if charge != np.sum(formal_charges):
                msg = (
                    f"Charge ({charge}) is different from sum of partial "
                    f"charges ({np.sum(formal_charges)})"
                )
                raise Exception(msg)
        if multiplicity is None:
            multiplicity = 1

        # Store conformers
        self.conformers: list[Conformer] = []
        self._add_conformers(conformer_coordinates, energies, degeneracies, properties)

        # Set up attributes
        self.charge = charge
        self.formal_charges = np.array(formal_charges)
        self.multiplicity = multiplicity
        if connectivity_matrix is not None:
            self.connectivity_matrix = np.array(connectivity_matrix)
        else:
            self.connectivity_matrix = None
        self.mol = None
        self.ref_cip_label = ref_cip_label

    def add_conformers(
        self,
        conformer_coordinates: ArrayLike3D,
        energies: ArrayLike1D | None = None,
        degeneracies: ArrayLike1D | None = None,
        properties: Mapping[str, ArrayLike1D] | None = None,
    ) -> ConformerEnsemble:
        """Add conformer to ensemble.

        Args:
            conformer_coordinates: Conformer coordinates (Å)
            energies: Energies (a.u.)
            degeneracies: Degeneracies
            properties: Conformer properties.

        Returns:
            self: Self
        """
        self._add_conformers(conformer_coordinates, energies, degeneracies, properties)

        return self

    def add_inverted(self) -> ConformerEnsemble:
        """Add inverted images of all conformers.

        Scrambles stereochemistry and leads to redundant conformers so use with
        care and prune based on RMSD as postprocessing.

        Returns:
            self: Self
        """
        conformers = self.conformers
        conformer_coordinates: list[Array2DFloat] = []
        energies: list[float | None] = []
        for conformer in conformers:
            coordinates = conformer.coordinates * np.array([-1, -1, -1])
            conformer_coordinates.append(coordinates)
            energies.append(conformer.energy)

        if not all([isinstance(energy, float) for energy in energies]):
            self._add_conformers(conformer_coordinates)
        else:
            energies_ = cast("list[float]", energies)
            self.add_conformers(conformer_coordinates, energies_)

        return self

    @requires_dependency([Import(module="rdkit.Chem", item="AllChem")], globals())
    def align_conformers(self) -> ConformerEnsemble:
        """Align conformers with RDKit."""
        self.update_mol()
        AllChem.AlignMolConformers(self.mol)

        return self

    def boltzmann_average_dT(
        self, property_name: str, temperature: float = 298.15
    ) -> float:
        """Calculate temperature derivative of Boltzmann average of property.

        Args:
            property_name: Name of property
            temperature: Temperature (K)

        Returns:
            derivative: Derivative of Boltzmann average
        """
        energies = self.get_energies()
        properties = self.get_properties()[property_name]
        derivative = boltzmann_average_dT(properties, energies, temperature=temperature)

        return derivative

    def boltzmann_statistic(
        self, property_name: str, temperature: float = 298.15, statistic: str = "avg"
    ) -> float:
        """Calculate Boltzmann staistic of property over ensemble.

        Args:
            property_name: Name of property
            statistic: Boltzmann statistic: 'avg', 'var' or 'std'
            temperature: Temperature (K)

        Returns:
            statistic: Boltzmann statistic
        """
        energies = self.get_energies()
        properties = self.get_properties()[property_name]
        statistic = boltzmann_statistic(
            properties, energies, temperature=temperature, statistic=statistic
        )

        return statistic

    def boltzmann_weights(self, temperature: float = 298.15) -> Array1DFloat:
        """Calculate Boltzmann weights for ensemble.

        Args:
            temperature: Temperature (K)

        Returns:
            weights: Conformer weights (normalized to unity)
        """
        energies = self.get_energies()
        weights = boltzmann_weights(energies, temperature)

        return weights

    def detect_enantiomers(
        self, thres: float = 0.01, method: str = "rdkit", include_hs: bool = False
    ) -> dict[int, list[int]]:
        """Detect enantiomers in ensemble.

        Args:
            thres: RMSD threshold for detecting enantiomers in terms of coordinates
            method: RMSD calculation method: 'obrms-batch', 'obrms-iter', openbabel' or
                'spyrmsd'
            include_hs: Whether to include H atoms when determining enantiomers

        Returns:
            enantiomers: Mapping of enantiomer with conformer id as keys and enantiomer
                ids as values.
        """
        # Add set of inverted conformers
        n_conformers = self.n_conformers
        self.add_inverted()

        # Map conformers to enantiomers
        enantiomer_map: dict[int, set[int]] = {i: set() for i in range(n_conformers)}
        rmsds = self.get_rmsd(method=method, include_hs=include_hs, symmetry=False)
        for i in range(n_conformers):
            # Do test that inverted conformer should have 0 RMSD to be
            # enantiomer
            test = np.where(rmsds[i, n_conformers:] < thres)[0]
            if len(test) > 0:
                if i not in test:
                    enantiomer_map[i].update(test)
                    for j in test:
                        enantiomer_map[j].add(i)
        enantiomers = {key: list(value) for key, value in enantiomer_map.items()}

        # Reset number of conformers
        self.conformers = self.conformers[:n_conformers]

        return enantiomers

    @classmethod
    def from_crest(cls: Type[T], path: str | PathLike) -> T:
        """Generate conformer ensemble from CREST output.

        Args:
            path: Path to CREST folder

        Returns:
            ce: Conformer ensemble object.
        """
        cp = CrestParser(path)
        ce = cls(
            cp.elements,
            cp.conformer_coordinates,
            energies=cp.energies,
            degeneracies=cp.degeneracies,
        )

        return ce

    @classmethod
    def from_rdkit(cls: Type[T], *args: Any, **kwargs: Any) -> T:
        """Generate conformer ensemble from RDKit.

        See the documentation for the function conformers_from_rdkit for
        more information.

        Args:
            *args: Positional arguments
            **kwargs: Keyword arguments

        Returns:
            ce: Conformer ensemble object.
        """
        # Run RDKit conformer search and generate ensemble.
        (
            elements,
            conformer_coordinates,
            energies,
            connectivity_matrix,
            charges,
            mol,
        ) = conformers_from_rdkit(*args, **kwargs)
        ce = cls(
            elements,
            conformer_coordinates,
            energies=energies,
            connectivity_matrix=connectivity_matrix,
            formal_charges=charges,
        )
        ce.mol = mol
        ce.set_multiplicity_from_mol()

        # Set reference CIP label if enantiomerically pure.
        cip_labels = ce.get_cip_labels()
        if len(set(cip_labels)) == 1:
            ce.ref_cip_label = cip_labels[0]

        return ce

    @classmethod
    def from_ob_ga(
        cls: Type[T],
        *args: Any,
        generate_rdkit_mol: bool = False,
        update_charges: bool = True,
        update_connectivity: bool = True,
        update_multiplicity: bool = True,
        **kwargs: Any,
    ) -> T:
        """Generate conformer ensemble from OpenBabel with GA method.

        See the documentation for the function conformers_from_ob_ga
        for more information.

        Args:
            *args: Positional arguments for conformers_from_ob_ga
            generate_rdkit_mol: Generate RDKit mol object for ensemble
            update_charges: Update formal charges from generated RDKit Mol object. Only
                used if generate_rdkit_mol is True.
            update_connectivity: Update connectivity from generated RDKit Mol object.
                Only used if generate_rdkit_mol is True.
            update_multiplicity: Update multiplicity from generated RDKit Mol object.
                Only used if generate_rdkit_mol is True.
            **kwargs: Keyword arguments for conformers_from_ob_ga

        Returns:
            ce: Conformer ensemble.
        """
        # Run Openbabel conformer search and generate ensemble.
        (
            elements,
            conformer_coordinates,
            connectivity_matrix,
            charges,
            ob_mol,
        ) = conformers_from_ob_ga(*args, **kwargs)
        ce = cls(
            elements,
            conformer_coordinates,
            connectivity_matrix=connectivity_matrix,
            formal_charges=charges,
        )

        # Generate RDKit mol object and CIP labels
        if generate_rdkit_mol:
            ce.generate_mol(
                update_charges=update_charges,
                update_connectivity=update_connectivity,
                update_multiplicity=update_multiplicity,
            )

            cip_labels = ce.get_cip_labels()
            if len(set(cip_labels)) == 1:
                ce.ref_cip_label = cip_labels[0]

        return ce

    @classmethod
    def from_ob_ff(
        cls: Type[T],
        *args: Any,
        generate_rdkit_mol: bool = False,
        update_charges: bool = True,
        update_connectivity: bool = True,
        update_multiplicity: bool = True,
        **kwargs: Any,
    ) -> T:
        """Generate conformer ensemble from OpenBabel with FF method.

        See the documentation for the function conformers_from_ob_ff
        for more information.

        Args:
            *args: Positional arguments for conformers_from_ob_ga
            generate_rdkit_mol: Generate RDKit mol object for ensemble
            update_charges: Update formal charges from generated RDKit Mol object. Only
                used if generate_rdkit_mol is True.
            update_connectivity: Update connectivity from generated RDKit Mol object.
                Only used if generate_rdkit_mol is True.
            update_multiplicity: Update multiplicity from generated RDKit Mol object.
                Only used if generate_rdkit_mol is True.
            **kwargs: Keyword arguments for conformers_from_ob_ga

        Returns:
            ce: Conformer ensemble.
        """
        # Run Openbabel conformer search and generate ensemble.
        (
            elements,
            conformer_coordinates,
            connectivity_matrix,
            charges,
            ob_mol,
        ) = conformers_from_ob_ff(*args, **kwargs)
        ce = cls(
            elements,
            conformer_coordinates,
            connectivity_matrix=connectivity_matrix,
            formal_charges=charges,
        )

        # Generate RDKit mol object and CIP labels
        if generate_rdkit_mol:
            ce.generate_mol(
                update_charges=update_charges,
                update_connectivity=update_connectivity,
                update_multiplicity=update_multiplicity,
            )

            cip_labels = ce.get_cip_labels()
            if len(set(cip_labels)) == 1:
                ce.ref_cip_label = cip_labels[0]

        return ce

    @requires_dependency([Import(module="rdkit", item="Chem")], globals())
    def generate_mol(
        self,
        update_charges: bool = True,
        update_connectivity: bool = True,
        update_multiplicity: bool = True,
    ) -> None:
        """Generate RDKit Mol object.

        Args:
            update_charges: Update formal charges from generated RDKit Mol object
            update_connectivity: Update connectivity from generated RDKit Mol object
            update_multiplicity: Update multiplicity from generated RDKit Mol object
        """
        mol = _get_rdkit_mol(
            self.elements,
            self.get_coordinates(),
            self.connectivity_matrix,
            self.formal_charges,
        )
        self.mol = mol

        if update_charges:
            self.formal_charges = np.array(
                [atom.GetFormalCharge() for atom in mol.GetAtoms()]
            )
        if update_connectivity:
            self.connectivity_matrix = np.array(
                Chem.GetAdjacencyMatrix(mol, useBO=True)
            )
        if update_multiplicity:
            self.set_multiplicity_from_mol()

    @requires_dependency([Import(module="rdkit", item="Chem")], globals())
    def get_cip_labels(self) -> list[tuple[str, ...]]:
        """Generate tuples of CIP labels for conformer.

        Returns:
            cip_labels: Tuples of CIP labels for each conformer.
        """
        # Update RDKit Mol object with current conformers.
        self.update_mol()

        # Generate CIP labels with new RDKit
        cip_labels = []
        mol = self.mol
        for i in range(mol.GetNumConformers()):
            Chem.AssignStereochemistryFrom3D(mol, i)
            labels = []
            for atom in mol.GetAtoms():
                if atom.HasProp("_CIPCode"):
                    label = atom.GetProp("_CIPCode")
                else:
                    label = ""
                labels.append(label)
            cip_labels.append(tuple(labels))

        return cip_labels

    def get_coordinates(self) -> Array3DFloat:
        """Get conformer coordinates.

        Returns:
            conformer_coordinates: Conformer coordinates (Å)
        """
        conformer_coordinates: Array3DFloat = np.array(
            [conformer.coordinates for conformer in self.conformers]
        )

        return conformer_coordinates

    def get_degeneracies(self) -> Array1DInt:
        """Get conformer degeneracies.

        Returns:
            degeneriacies: Degeneracies
        """
        degeneracies: Array1DInt = np.array(
            [conformer.degeneracy for conformer in self.conformers]
        )

        return degeneracies

    def get_energies(self) -> Array1DFloat:
        """Get conformer energies.

        Returns:
            energies: Energy (a.u.)
        """
        energies: Array1DFloat = np.array(
            [conformer.energy for conformer in self.conformers]
        )
        return energies

    def get_properties(self) -> dict[str, Array1DFloat]:
        """Get conformer properties.

        Returns:
            properties: Conformer properties
        """
        properties: dict[str, list[float]] = {}
        for conformer in self.conformers:
            for key, value in conformer.properties.items():
                properties.setdefault(key, []).append(value)
        properties: dict[str, Array1DFloat] = {
            key: np.array(value) for key, value in properties.items()
        }

        return properties

    def get_relative_energies(
        self, unit: str = "kcal/mol", relative: bool = True
    ) -> Array1DFloat:
        """Get conformer energies with choice of units and reference value.

        Args:
            unit: Unit of returned energies: 'hartree', 'kcal/mol' or 'kJ/mol'.
            relative: Return energies relative to lowest-energy conformer.

        Returns:
            energies: Conformer energies.
        """
        energies = self.get_energies()

        # Adjust unit
        if unit.lower() == "hartree":
            pass
        if unit.lower() == "kcal/mol":
            energies *= HARTREE_TO_KCAL
        elif unit.lower() == "kj/mol":
            energies *= HARTREE_TO_KJ

        if relative:
            energies -= np.min(energies)

        return energies

    def get_rmsd(
        self,
        i_s: ArrayLike1D | None = None,
        j_s: ArrayLike1D | None = None,
        include_hs: bool = False,
        symmetry: bool = False,
        method: str = "rdkit",
    ) -> Array2DFloat:
        """Get RSMD between conformers.

        For very small systems 'openbabel' or 'spyrmsd' work well. For larger systems a
        significant speed-up is attained with 'rdkit', 'obrms-batch' or 'obrms-iter'.

        Args:
            i_s: Indices of conformers
            j_s: Indices of conformers
            include_hs: Whether to include H atoms in RMSD calculation. Ignored for
                'obrms-iter' and 'obrms-batch' that only use heavy atoms.
            symmetry: Consider symmetry (requires connectivity matrix). Ignored for
                'obrms-iter' and 'obrms-batch' that always use symmetry.
            method: RMSD calculation method: 'obrms-batch', 'obrms-iter', 'openbabel',
                'rdkit' or 'spyrmsd'.

        Returns:
            rmsds: RSMDs (Å)

        Raises:
            ValueError: If method not supported.
        """
        i_s_: np.ndarray
        if i_s is None:
            i_s_ = np.arange(1, len(self.conformers) + 1)
        else:
            i_s_ = np.array(i_s)
        i_s = i_s_
        j_s_: np.ndarray
        if j_s is None:
            j_s_ = np.arange(1, len(self.conformers) + 1)
        else:
            j_s_ = np.array(j_s)
        j_s = j_s_

        if method == "obrms-batch":
            rmsds = self._get_rmsd_obrms_batch(i_s, j_s)
        elif method == "obrms-iter":
            rmsds = self._get_rmsd_obrms_iter(i_s, j_s)
        elif method == "openbabel":
            rmsds = self._get_rmsd_openbabel(i_s, j_s, include_hs, symmetry)
        elif method == "rdkit":
            rmsds = self._get_rmsd_rdkit(i_s, j_s, include_hs)
        elif method == "spyrmsd":
            rmsds = self._get_rmsd_spyrmsd(i_s, j_s, include_hs, symmetry)
        else:
            raise ValueError("Method not supported.")
        return rmsds

    @property
    def n_conformers(self) -> int:
        """Number of conformers."""
        n_conformers = len(self.conformers)

        return n_conformers

    def optimize_qc_engine(
        self,
        ids: Sequence[int] | None = None,
        program: str | None = None,
        model: dict[str, Any] | None = None,
        keywords: dict[str, Any] | None = None,
        local_options: dict[str, Any] | None = None,
        procedure: str = "berny",
    ) -> ConformerEnsemble:
        """Optimize conformers with QCEngine interface.

        Args:
            ids: Conformer indices (1-indexed). If None, all are optimized
            program: QCEngine program
            model: QCEngine model
            keywords: QCEngine keywords
            local_options: QCEngine local options
            procedure: QCEngine procedure

        Returns:
            self: Self

        Raises:
            Exception: When RDKit requested without formal charges.
        """
        if program is not None:
            if np.any(self.formal_charges != 0) and (program.lower() == "rdkit"):
                raise Exception("QCEngine using RDKit does not work with charges.")

        # Set defaults
        if model is None:
            model = {}
        if keywords is None:
            keywords = {}
        if local_options is None:
            local_options = {}
        if ids is not None:
            conformers = [self.conformers[i] for i in ids]
        else:
            conformers = self.conformers

        for conformer in conformers:
            opt_coordinates, energies = optimize_qc_engine(
                self.elements,
                conformer.coordinates,
                charge=self.charge,
                multiplicity=self.multiplicity,
                connectivity_matrix=self.connectivity_matrix,
                program=program,
                model=model,
                keywords=keywords,
                local_options=local_options,
                procedure=procedure,
            )
            conformer.coordinates = opt_coordinates
            conformer.energy = energies[-1]

        return self

    def condense_enantiomeric(self, thres: float | None = None) -> ConformerEnsemble:
        """Condense enantiomers into single enantiomer per pair.

        Args:
            thres: RMSD threshold for assessing enantiomers.

        Returns:
            self: Self

        Raises:
            Exception: If molecule is chiral.
        """
        cip_labels = self.get_cip_labels()
        if "".join(["".join(labels) for labels in cip_labels]) != "":
            raise Exception("Chiral molecule. Not safe to condense.")

        enantiomers = self.detect_enantiomers(thres=0.01)
        to_keep = []
        to_remove = []
        for key, value in enantiomers.items():
            if key not in to_remove:
                conformer = self.conformers[key]
                conformer.degeneracy = len(value) + 1
                to_keep.append(conformer)
                to_remove.extend(value)

        self.conformers = to_keep

        return self

    def prune_enantiomers(
        self, keep: str = "original", ref_label: tuple[str, ...] | None = None
    ) -> ConformerEnsemble:
        """Prune conformers so that only one enantiomer is present in the ensemble.

        Args:
            keep: Which enantiomer to keep: 'original', 'most common' or 'specified'.
                Choice of 'original' requires that the ref_cip_label attribute is set.
                Choice of 'specified' requires ref_label to be given.
            ref_label: Reference CIP labels for all atoms

        Returns:
            self: Self
        """
        cip_labels = self.get_cip_labels()

        # Set up reference label
        if keep == "original":
            ref_label = self.ref_cip_label
            if ref_label is None:
                keep = "most common"
        elif keep == "specified":
            pass
        if keep == "most common":
            counter: Counter[tuple[str, ...]] = Counter(cip_labels)
            ref_label = counter.most_common(n=1)[0][0]

        # Prune conformers
        to_keep = []
        for i, cip_label in enumerate(cip_labels):
            if cip_label == ref_label:
                to_keep.append(i)
        self.conformers = [self.conformers[i] for i in to_keep]

        return self

    def prune_energy(
        self, threshold: float = 3.0, unit: str = "kcal/mol"
    ) -> ConformerEnsemble:
        """Prune conformers based on energy compared to minimum energy conformer.

        Args:
            threshold: Energy threshold for pruning
            unit: Unit for energy threshold 'hartree', 'kcal/mol' or 'kJ/mol'

        Returns:
            self: Self
        """
        # Convert threshold to Hartree.
        if unit.lower() == "hartree":
            pass
        if unit.lower() == "kcal/mol":
            threshold *= KCAL_TO_HARTREE
        elif unit.lower() == "kJ/mol":
            threshold *= KJ_TO_HARTREE

        # Prune conformers.
        energies = self.get_energies()
        energies -= np.min(energies)
        remove_list = np.where(energies > threshold)[0]

        for i in reversed(remove_list):
            del self.conformers[i]

        return self

    def prune_rmsd(
        self,
        thres: float = 0.35,
        include_hs: bool = False,
        symmetry: bool = False,
        method: str = "rdkit",
    ) -> ConformerEnsemble:
        """Prune conformers based on RMSD.

        Args:
            thres: Threshold for RSMD pruning (Å)
            include_hs: Whether to include H atoms in RMSD calculation. Ignored for
                'obrms-iter' and 'obrms-batch' which only use heavy atoms.
            symmetry: Consider symmetry (requires connectivity matrix). Ignored for
                'obrms-iter' and 'obrms-batch' which always use symmetry.
            method: RMSD calculation method: 'obrms-batch', 'obrms-iter', 'openbabel',
                'rdkit' or 'spyrmsd'

        Returns:
            self: Self
        """
        # Select conformers to keep
        candidates = np.arange(len(self.conformers))
        keep_list = []

        # Prune conformers iteratively if possible to reduce number of RMSD
        # evaluations. For large systems obrms might be faster in batch.
        if method in ("obrms-iter", "openbabel", "spyrmsd", "rdkit"):
            while len(candidates) > 0:
                keeper = candidates[0]
                keep_list.append(keeper)
                rmsd = self.get_rmsd(
                    [keeper + 1],
                    candidates + 1,
                    include_hs=include_hs,
                    symmetry=symmetry,
                    method=method,
                )
                candidates = candidates[rmsd[0] > thres]
        elif method == "obrms-batch":
            rmsds = self.get_rmsd(
                include_hs=include_hs, symmetry=symmetry, method=method
            )
            working_array = rmsds
            while len(working_array) > 0:
                keeper = candidates[0]
                keep_list.append(keeper)
                rmsd = working_array[0]
                mask = rmsd > thres
                candidates = candidates[mask]
                working_array = working_array[mask, :][:, mask]

        # Update conformer list
        self.conformers = [
            conformer for i, conformer in enumerate(self.conformers) if i in keep_list
        ]

        return self

    def set_coordinates(self, conformer_coordinates: ArrayLike3D) -> ConformerEnsemble:
        """Set conformer coordinates.

        Args:
            conformer_coordinates: Conformer coordinates (Å)

        Returns:
            self: Self

        Raises:
            ValueError: When number of conformer coordinates is different from number of
                conformers
        """
        conformer_coordinates: Array3DFloat = np.array(conformer_coordinates)
        if len(conformer_coordinates) != self.n_conformers:
            msg = (
                f"Number of coordinates ({len(conformer_coordinates)}) "
                f"!= number of conformers ({self.n_conformers})."
            )
            raise ValueError(msg)
        for conformer, coordinates in zip(self.conformers, conformer_coordinates):
            conformer.coordinates = np.array(coordinates)

        return self

    def set_degeneracies(self, degeneracies: ArrayLike1D) -> ConformerEnsemble:
        """Set degeneracies.

        Args:
            degeneracies: Degeneracies

        Returns:
            self: Self

        Raises:
            ValueError: When number of degeneracies is different from number of
                conformers
        """
        degeneracies: Array1DInt = np.array(degeneracies)
        if len(degeneracies) != self.n_conformers:
            msg = (
                f"Number of degeneracies ({len(degeneracies)}) "
                f"!= number of conformers ({self.n_conformers})."
            )
            raise ValueError(msg)
        for conformer, degeneracy in zip(self.conformers, degeneracies):
            conformer.degeneracy = degeneracy

        return self

    def set_energies(self, energies: ArrayLike1D) -> ConformerEnsemble:
        """Set energies.

        Args:
            energies: Energy (a.u.)

        Returns:
            self: Self

        Raises:
            ValueError: When number of energies is different from number of
                conformers
        """
        energies: Array1DFloat = np.array(energies)
        if len(energies) != self.n_conformers:
            msg = (
                f"Number of energies ({len(energies)}) != number of "
                f"conformers ({self.n_conformers})."
            )
            raise ValueError(msg)
        for conformer, energy in zip(self.conformers, energies):
            conformer.energy = energy

        return self

    @requires_dependency([Import(module="rdkit.Chem", item="Descriptors")], globals())
    def set_multiplicity_from_mol(self) -> ConformerEnsemble:
        """Sets multiplicity based on unpaired electrons in Mol object."""
        num_radical = Descriptors.NumRadicalElectrons(self.mol)

        # Assume maximum spin pairing
        if (num_radical % 2) == 0:
            multiplicity = 1
        else:
            multiplicity = 2

        self.multiplicity = multiplicity

        return self

    def set_properties(self, key: str, values: Iterable[float]) -> ConformerEnsemble:
        """Set conformer properties.

        Args:
            key: Name of property
            values: Property values

        Returns:
            self: Self
        """
        for conformer, value in zip(self.conformers, values):
            conformer.properties[key] = value

        return self

    def sort(self) -> ConformerEnsemble:
        """Sort conformers based on energy."""
        energies = [conformer.energy for conformer in self.conformers]
        if not all([isinstance(energy, float) for energy in energies]):
            raise ValueError("Not all conformers have energies.")
        energies = cast("list[float]", energies)
        indices = np.argsort(energies)
        self.conformers = [self.conformers[i] for i in indices]

        return self

    def sp_qc_engine(
        self,
        ids: Sequence[int] | None = None,
        program: str = "xtb",
        model: dict[str, Any] | None = None,
        keywords: dict[str, Any] | None = None,
        local_options: dict[str, Any] | None = None,
    ) -> ConformerEnsemble:
        """Calculate conformer energies with QCEngine interface.

        Args:
            ids: Conformer indices (1-indexed). If None, all are calculated.
            program: QCEngine program
            model: QCEngine model
            keywords: QCEngine keywords
            local_options: QCEngine local options

        Returns:
            self: Self

        Raises:
            Exception: When trying to use RDKit with formal charges.
        """
        if np.any(self.formal_charges != 0) and (program.lower() == "rdkit"):
            raise Exception("QCEngine using RDKit does not work with charges.")

        # Set defaults
        if model is None:
            model = {"method": "GFN2-xTB"}
        if keywords is None:
            keywords = {}
        if local_options is None:
            local_options = {}
        if ids is not None:
            conformers = [self.conformers[i] for i in ids]
        else:
            conformers = self.conformers

        for conformer in conformers:
            energy = sp_qc_engine(
                self.elements,
                conformer.coordinates,
                charge=self.charge,
                multiplicity=self.multiplicity,
                connectivity_matrix=self.connectivity_matrix,
                program=program,
                model=model,
                keywords=keywords,
                local_options=local_options,
            )
            conformer.energy = energy

        return self

    def write_xyz(
        self,
        file: str | PathLike,
        ids: Iterable[int] | None = None,
        unit: str = "kcal/mol",
        relative: bool = True,
        separate: bool = False,
        n_decimals: int = 3,
    ) -> None:
        """Write conformers to xyz file.

        Args:
            file: Filename or path object. Needs filename if `separate=True`
            ids: Conformer indices (1-indexed)
            unit: Output unit for energies in xyz comment field: 'hartree', 'kcal/mol',
                'kJ/mol'
            relative: Whether to give energies relative to lowest energy conformer
            separate: Whether to write conformers to separate xyz files
            n_decimals: Number of decimals for energies

        Raises:
            TypeError: When separate=True and file is not str
        """
        ids_: np.ndarray
        if ids is None:
            ids_ = np.arange(len(self.conformers))
        else:
            ids_ = np.array(ids) - 1
        ids = ids_

        # Retrieve symbols, coordinates and energies
        symbols = convert_elements(self.elements, output="symbols")
        conformer_coordinates = self.get_coordinates()[ids]
        energies = self.get_relative_energies(unit=unit, relative=relative)[ids].round(
            n_decimals
        )

        # Write conformers
        if separate:
            if not isinstance(file, str):
                raise TypeError("file must be str when separate=True")
            for i, coordinates, energy in zip(ids, conformer_coordinates, energies):
                conf_filename = file.split(".")[0] + f"_{i + 1}.xyz"
                write_xyz(conf_filename, symbols, coordinates, comments=[energy])
        else:
            write_xyz(file, symbols, conformer_coordinates, comments=energies)

    def _add_conformers(
        self,
        conformer_coordinates: ArrayLike3D,
        energies: ArrayLike1D | None = None,
        degeneracies: ArrayLike1D | None = None,
        properties: Mapping[str, ArrayLike1D] | None = None,
    ) -> None:
        conformer_coordinates: Array3DFloat = np.array(conformer_coordinates)
        n_conformers = len(conformer_coordinates)

        energies_: np.ndarray
        if energies is None:
            energies_ = np.full(n_conformers, np.nan)
        else:
            energies_ = np.array(energies)
        energies = energies_

        degeneracies_: np.ndarray
        if degeneracies is None:
            degeneracies_ = np.ones(n_conformers)
        else:
            degeneracies_ = np.array(degeneracies)
        degeneracies = degeneracies_

        for coordinates, energy, degeneracy in zip(
            conformer_coordinates, energies, degeneracies
        ):
            conformer = Conformer(self.elements, coordinates, energy, degeneracy)
            self.conformers.append(conformer)
        if properties is not None:
            properties_: dict[str, Array1DFloat] = {
                key: np.array(value) for key, value in properties.items()
            }
            for key, value in properties_.items():
                self.set_properties(key, value)

    @requires_executable(["obrms"])
    def _get_rmsd_obrms_batch(self, i_s: Array1DInt, j_s: Array1DInt) -> Array2DFloat:
        """Calculate RMSD with obrms in batch mode.

        First calculates matrix of all pairwise RMSDs and then takes those of interest.

        Args:
            i_s: Conformer indices (1-indexed)
            j_s: Conformer indices (1-indexed)

        Returns:
            rmsds: RMSDs (Å)
        """
        with tempfile.NamedTemporaryFile(suffix=".xyz") as ref_file:
            p_ref = Path(ref_file.name)
            p_ref.unlink()
            self.write_xyz(p_ref, unit="hartree", relative=False)
            process = subprocess.run(
                f"obrms {p_ref.as_posix()} --cross " "--minimize".split(" "),
                capture_output=True,
            )
        result = np.genfromtxt(process.stdout.splitlines(), delimiter=",")
        # Reshape array to 2D if 1D
        if len(result.shape) == 1:
            result = result.reshape(1, -1)
        rmsds = result[:, 1:]

        rmsds: Array2DFloat = rmsds[i_s - 1, :][:, j_s - 1]

        return rmsds

    @requires_executable(["obrms"])
    def _get_rmsd_obrms_iter(self, i_s: Array1DInt, j_s: Array1DInt) -> Array2DFloat:
        """Calculate RMSD with obrms.

        Does iterative row-wise mode for heavy atoms and without symmetry.

        Args:
            i_s: Conformer indices (1-indexed)
            j_s: Conformer indices (1-indexed)

        Returns:
            rmsds: Conformer RMSDs (Å)
        """
        rmsds = []
        for i in i_s:
            with tempfile.NamedTemporaryFile(
                suffix=".xyz"
            ) as ref_file, tempfile.NamedTemporaryFile(suffix=".xyz") as test_file:
                p_ref = Path(ref_file.name)
                p_ref.unlink()
                p_test = Path(test_file.name)
                p_test.unlink()
                self.write_xyz(p_ref, ids=j_s)
                self.write_xyz(p_test, ids=[i])
                process = subprocess.run(
                    f"obrms {p_ref.as_posix()} "
                    f"{p_test.as_posix()} --minimize".split(" "),
                    capture_output=True,
                )
            row_rmsds = np.genfromtxt(process.stdout.splitlines(), usecols=(-1))  # type: ignore
            rmsds.append(row_rmsds)
        rmsds: Array2DFloat = np.vstack(rmsds)

        return rmsds

    @requires_dependency([Import(module="openbabel.openbabel", alias="ob")], globals())
    def _get_rmsd_openbabel(
        self, i_s: Array1DInt, j_s: Array1DInt, include_hs: bool, symmetry: bool
    ) -> Array2DFloat:
        """Calculate RMSD row-wise with openbabel python interface.

        Args:
            i_s: Conformer indices (1-indexed)
            j_s: Conformer indices (1-indexed)
            include_hs: Whether to include H atoms in calculation
            symmetry: Whether to consider symmetry

        Returns:
            rmsds: Conformer RMSDs (Å)
        """
        rmsds = []
        for i in i_s:
            conformer_1 = self.conformers[i - 1]
            ob_mol_1 = _get_ob_mol(
                self.elements, conformer_1.coordinates, self.connectivity_matrix
            )
            align = ob.OBAlign(include_hs, symmetry)
            align.SetRefMol(ob_mol_1)
            rmsds_row = []
            for j in j_s:
                conformer_2 = self.conformers[j - 1]
                ob_mol_2 = _get_ob_mol(
                    self.elements, conformer_2.coordinates, self.connectivity_matrix
                )
                align.SetTargetMol(ob_mol_2)
                align.Align()
                rmsd = align.GetRMSD()
                rmsds_row.append(rmsd)
            rmsds.append(rmsds_row)
        rmsds: Array2DFloat = np.array(rmsds)

        return rmsds

    @requires_dependency(
        [
            Import(module="rdkit", item="Chem"),
            Import(module="rdkit.Chem", item="AllChem"),
        ],
        globals(),
    )
    def _get_rmsd_rdkit(
        self, i_s: Array1DInt, j_s: Array1DInt, include_hs: bool
    ) -> Array2DFloat:
        """Calculate RMSD row-wise with RDKit.

        Args:
            i_s: Conformer indices (1-indexed)
            j_s: Conformer indices (1-indexed)
            include_hs: Whether to include H atoms in calculation

        Returns:
            rmsds: Conformer RMSDs (Å)
        """
        # Update mol object from conformers and make copy.
        self.update_mol()

        # Construct atom list for rmsd: heavy atoms or all atoms
        if include_hs:
            atom_ids = [atom.GetIdx() for atom in self.mol.GetAtoms()]
        else:
            atom_ids = [
                atom.GetIdx()
                for atom in self.mol.GetAtoms()
                if atom.GetAtomicNum() != 1
            ]

        # Calculated RMSD row-wise with RDKit
        rmsds = []
        conformers = list(self.mol.GetConformers())
        for i in i_s:
            ref_mol = Chem.Mol(self.mol)
            ref_conformer = conformers[i - 1]
            ref_mol.RemoveAllConformers()
            ref_mol.AddConformer(ref_conformer)
            for j in j_s:
                conformer = conformers[j - 1]
                ref_mol.AddConformer(conformer)
            rmsds_row: list[float] = []
            AllChem.AlignMolConformers(ref_mol, atomIds=atom_ids, RMSlist=rmsds_row)
            rmsds.append(rmsds_row)
        rmsds: Array2DFloat = np.array(rmsds)

        return rmsds

    @requires_dependency([Import("spyrmsd"), Import("spyrmsd.rmsd")], globals())
    def _get_rmsd_spyrmsd(
        self, i_s: Array1DInt, j_s: Array1DInt, include_hs: bool, symmetry: bool
    ) -> Array2DFloat:
        """Calculate RMSD row-wise with spyrmsd.

        Args:
            i_s: Conformer indices (1-indexed)
            j_s: Conformer indices (1-indexed)
            include_hs: Whether to include H atoms in calculation
            symmetry: Whether to consider symmetry

        Returns:
            rmsds: Conformer RMSDs (Å)
        """
        # Construct mask for H atoms
        if not include_hs:
            mask = self.elements != 1
        else:
            mask = np.ones(len(self.elements), dtype=bool)

        # Calculate RMSD row-wise with spyrmsd
        rmsds: list[Array1DFloat] = []
        for i in i_s:
            ref_coordinates = self.conformers[i - 1].coordinates[mask]
            if symmetry:
                test_coordinates = [
                    self.conformers[j - 1].coordinates[mask] for j in j_s
                ]
                row_rmsds = spyrmsd.rmsd.symmrmsd(
                    ref_coordinates,
                    test_coordinates,
                    self.elements[mask],
                    self.elements[mask],
                    self.connectivity_matrix[mask, :][:, mask],
                    self.connectivity_matrix[mask, :][:, mask],
                    center=True,
                    minimize=True,
                )
            else:
                row_rmsds = []
                for j in j_s:
                    test_coordinates = self.conformers[j - 1].coordinates[mask]
                    rmsd = spyrmsd.rmsd.rmsd(
                        ref_coordinates,
                        test_coordinates,
                        self.elements[mask],
                        self.elements[mask],
                        center=True,
                        minimize=True,
                    )
                    row_rmsds.append(rmsd)
            rmsds.append(np.array(row_rmsds))
        rmsds: Array2DFloat = np.vstack(rmsds)

        return rmsds

    def update_mol(self) -> ConformerEnsemble:
        """Update Mol object with conformers."""
        self.mol.RemoveAllConformers()
        conformer_coordinates = self.get_coordinates()
        _add_conformers_to_mol(self.mol, conformer_coordinates)

        return self

    def __copy__(self) -> ConformerEnsemble:
        # Generate copy where conformers and mol object are shared with old
        # ensemble.
        cls = type(self)
        ce = cls(
            self.elements,
            connectivity_matrix=self.connectivity_matrix,
            charge=self.charge,
            multiplicity=self.multiplicity,
            formal_charges=self.formal_charges,
            ref_cip_label=self.ref_cip_label,
        )
        ce.mol = self.mol
        ce.conformers = self.conformers
        return ce

    def __deepcopy__(self, memo: dict[int, object]) -> ConformerEnsemble:
        # Generate copy where conformers and mol object are new
        cls = type(self)
        ce = cls(
            self.elements,
            conformer_coordinates=self.get_coordinates(),
            energies=self.get_energies(),
            connectivity_matrix=self.connectivity_matrix,
            degeneracies=self.get_degeneracies(),
            properties=self.get_properties(),
            charge=self.charge,
            multiplicity=self.multiplicity,
            formal_charges=self.formal_charges,
            ref_cip_label=self.ref_cip_label,
        )
        ce.mol = deepcopy(ce.mol, memo)
        return ce

    def __delitem__(self, index: int) -> None:
        del self.conformers[index]

    def __getitem__(
        self, index: slice | numbers.Integral
    ) -> ConformerEnsemble | Conformer:
        cls = type(self)
        if isinstance(index, slice):
            # Generate copy of ensemble with selected conformers.
            ce = copy(self)
            ce.conformers = ce.conformers[index]
            return ce
        elif isinstance(index, numbers.Integral):
            conformer = self.conformers[int(index)]
            return conformer
        else:
            raise TypeError(f"{cls.__name__} indices must be integers")

    def __len__(self) -> int:
        return len(self.conformers)

    def __repr__(self) -> str:
        n_conformers = len(self.conformers)
        return f"{self.__class__.__name__}({n_conformers!r} conformers)"


@requires_dependency(
    [
        Import(module="openbabel.openbabel", alias="ob"),
        Import(module="openbabel.pybel", alias="pybel"),
        Import("openbabel"),
    ],
    globals(),
)
def conformers_from_ob_ff(
    mol: str | ob.OBMol,
    n_conformers: int = 30,
    ff: str = "MMFF94",
    method: str = "systematic",
    rings: bool = False,
) -> tuple[Array1DInt, Array3DFloat, Array2DInt, Array1DInt, ob.OBMol]:
    """Generates conformers based on the force field algorithm in OpenBabel.

    Follows the recipe of the command line script obabel --conformer:
    https://github.com/openbabel/openbabel/blob/master/src/ops/conformer.cpp
    If an OBMol object with 3D coordinates is given, the conformer search will
    start from that structure.

    Args:
        mol: Molecule either as SMILES string or OBMol object.
        n_conformers: Maximum number of conformers
        ff: Force field supported by OpenBabel
        method: 'fast', random', 'systematic' or 'weighted'
        rings: Sample ring torsions.

    Returns:
        elements: Elements as atomic numbers
        conformer_coordinates: Conformer coordinates (Å)
        connectivity_matrix: Connectivity matrix
        charges: Formal charges
        mol: OBMol object
    """
    # Create 3D structure if not given
    try:
        py_mol = pybel.readstring("smi", mol)
        py_mol.make3D()
        ob_mol = py_mol.OBMol
    except TypeError:
        ob_mol = mol
        py_mol = pybel.Molecule(ob_mol)

    # Set up FF
    ff = ob.OBForceField_FindForceField(ff)
    ff.EnableCutOff(True)
    ff.SetVDWCutOff(10.0)
    ff.SetElectrostaticCutOff(20.0)
    ff.SetUpdateFrequency(10)

    # Do conformer search
    if method == "systematic":
        ff.SystematicRotorSearch(10, rings)
    elif method == "fast":
        ff.FastRotorSearch(True)
    elif method == "random":
        ff.RandomRotorSearch(n_conformers, 10, rings)
    elif method == "weighted":
        ff.WeightedRotorSearch(n_conformers, 10, rings)
    ff.GetConformers(ob_mol)

    # Extract information
    (
        elements,
        conformer_coordinates,
        connectivity_matrix,
        charges,
    ) = _extract_from_ob_mol(ob_mol)

    return (elements, conformer_coordinates, connectivity_matrix, charges, ob_mol)


@requires_dependency(
    [
        Import(module="openbabel.openbabel", alias="ob"),
        Import(module="openbabel.pybel", alias="pybel"),
        Import("openbabel"),
    ],
    globals(),
)
def conformers_from_ob_ga(  # noqa: C901
    mol: str | ob.OBMol,
    n_conformers: int | None = None,
    n_children: int | None = None,
    mutability: float | None = None,
    convergence: int | None = None,
    score: str = "rmsd",
    filter_method: str = "steric",
    cutoff: float = 0.8,
    vdw_factor: float = 0.5,
    check_hydrogens: bool = True,
) -> tuple[Array1DInt, Array3DFloat, Array2DInt, Array1DInt, ob.OBMol]:
    """Generates conformers based on the genetic algorithm in OpenBabel.

    Follows the recipe of the command line script obabel --conformer:
    https://github.com/openbabel/openbabel/blob/master/src/ops/conformer.cpp If an OBMol
    object with 3D coordinates is given, the conformer search will start from that
    structure.

    Args:
        mol: Molecule either as SMILES string or OBMol object
        n_conformers: Maximum number of conformers
        n_children: Number of children to generate for each parent
        mutability: Mutation frequency
        convergence: Number of identical generations before convergence is reached
        score: Scoring function: 'rmsd', 'min_rmsd', 'energy', 'min_energy'
        filter_method: Filtering algorithm: 'steric'
        cutoff: Absolute distance in Ånström below which atoms are considered to clash
        vdw_factor: Scale factor applied to van der Waals radii for detecting clashes
        check_hydrogens: Detect clashes with hydrogen atoms

    Returns:
        elements: Elements as atomic numbers
        conformer_coordinates: Conformer coordinates (Å)
        connectivity_matrix: Connectivity matrix
        charges: Formal charges
        mol: OBMol object
    """
    # Create 3D structure if not given
    try:
        py_mol = pybel.readstring("smi", mol)
        py_mol.make3D()
        ob_mol = py_mol.OBMol
    except TypeError:
        ob_mol = mol
        py_mol = pybel.Molecule(ob_mol)

    # Create search object and set parameters
    conf_search = ob.OBConformerSearch()
    conf_search.Setup(ob_mol)
    if n_conformers is not None:
        conf_search.SetNumConformers(n_conformers)
    if n_children is not None:
        conf_search.SetNumChildren(n_children)
    if convergence is not None:
        conf_search.SetConvergence(convergence)
    if mutability is not None:
        conf_search.SetMutability(mutability)
    if score is not None:
        # Scorers don't work with earlier versions of openbabel
        if not (parse(openbabel.__version__) > parse("3.1.0")):
            warnings.warn(
                "Scorer only works with openbabel version > 3.1.0. "
                "Proceeding without scorer.",
                stacklevel=2,
            )
        else:
            if score == "rmsd":
                scorer = ob.OBRMSDConformerScore()
            if score == "min_rmsd":
                scorer = ob.OBMinimizingRMSDConformerScore()
            if score == "energy":
                scorer = ob.OBEnergyConformerScore()
            if score == "min_energy":
                scorer = ob.OBMinimizingEnergyConformerScore()
            conf_search.SetScore(scorer)
    if filter_method == "steric":
        # Filters don't work with earlier versions of openbabel
        if not (parse(openbabel.__version__) > parse("3.1.0")):
            warnings.warn(
                "Filter only works with openbabel version > 3.1.0. "
                "Proceeding without filter.",
                stacklevel=2,
            )
        else:
            ob_filter = ob.OBStericConformerFilter(cutoff, vdw_factor, check_hydrogens)
            conf_search.SetFilter(ob_filter)

    # Do conformational search
    conf_search.Search()
    conf_search.GetConformers(ob_mol)

    # Extract information
    (
        elements,
        conformer_coordinates,
        connectivity_matrix,
        charges,
    ) = _extract_from_ob_mol(ob_mol)

    return (elements, conformer_coordinates, connectivity_matrix, charges, ob_mol)


@requires_dependency(
    [
        Import(module="rdkit", item="Chem"),
        Import(module="rdkit.Chem", item="AllChem"),
        Import(module="rdkit.Chem", item="rdDistGeom"),
    ],
    globals(),
)
def conformers_from_rdkit(  # noqa: C901
    mol: str | Chem.Mol,
    n_conformers: int | None = None,
    optimize: str | None = "MMFF94",
    version: int = 2,
    small_rings: bool = True,
    macrocycles: bool = True,
    random_seed: int | None = None,
    rmsd_thres: float | None = 0.35,
    rmsd_symmetry: bool = False,
    n_threads: int = 1,
) -> tuple[
    Array1DInt, Array3DFloat, Array1DFloat | None, Array2DFloat, Array1DInt, Chem.Mol
]:
    """Generates conformers for an RDKit mol object.

    Recipe based on J. Chem. Inf. Modeling 2012, 52, 1146.

    Args:
        mol: Molecule either as SMILES string or RDKit Mol object.
        n_conformers: Number of conformers to generate. The actual number
            of conformers returned may be lower due to RMSD pruning. If None,
            a reasonable number will be set depending on the number of
            rotatable bonds.
        optimize: Force field used for conformer optimization: 'MMFF94', 'MMFF94s' or
            'UFF'. If None, conformers are not optimized.
        version: Version of the experimental torsion-angle preferences
        small_rings: Whether to impose small ring torsion angle preferences
        macrocycles: Whether to impose macrocycle torsion angle preferences
        random_seed: Random seed for conformer generation
        rmsd_thres: Pruning RMSD threshold (Å)
        rmsd_symmetry: Whether to use symmetry for RMSD pruning
        n_threads: Number of threads

    Returns:
        elements: Atomic symbols
        conformer_coordinates: Coordinates for all conformers (Å)
        energies: Conformer energies (a.u.)
        connectivity_matrix: Connectivity matrix with bond orders
        charges: Formal charges
        mol: RDKit Mol object. Only returned when return_mol=True

    Raises:
        Exception: When force field not found
    """
    if optimize not in ("MMFF94", "MMFF94s", "UFF", None):
        raise Exception(
            f"Force field {optimize} not found. Choose one of" "MMFF94, MMFF94s, UFF."
        )
    # Generate mol object
    try:
        mol = Chem.MolFromSmiles(mol)
    except TypeError:
        pass
    mol: Chem.Mol = Chem.AddHs(mol)

    # If n_conformers is not set, set number of conformers based on number of
    # rotatable bonds
    if not n_conformers:
        n_rot_bonds = AllChem.CalcNumRotatableBonds(mol)

        if n_rot_bonds <= 7:
            n_conformers = 50
        elif n_rot_bonds >= 8 and n_rot_bonds <= 12:
            n_conformers = 200
        else:
            n_conformers = 300

    # Generate conformers
    if rmsd_thres is not None:
        rdkit_prune_rmsd = rmsd_thres
    else:
        rdkit_prune_rmsd = -1
    if random_seed is not None:
        rdkit_random_seed = random_seed
    else:
        rdkit_random_seed = -1

    params = AllChem.EmbedParameters()
    params.useSymmetryForPruning = rmsd_symmetry
    params.randomSeed = rdkit_random_seed
    params.useSmallRingTorsions = small_rings
    params.useMacrocycleTorsions = macrocycles
    params.ETversion = version
    params.pruneRmsThresh = rdkit_prune_rmsd
    params.numThreads = n_threads
    rdDistGeom.EmbedMultipleConfs(mol, n_conformers, params)

    # Optimize with force fields
    results = None
    if optimize is not None:
        if optimize == "MMFF94":
            results = AllChem.MMFFOptimizeMoleculeConfs(
                mol, mmffVariant="MMFF94", numThreads=n_threads
            )
        elif optimize == "MMFF94s":
            results = AllChem.MMFFOptimizeMoleculeConfs(
                mol, mmffVariant="MMFF94s", numThreads=n_threads
            )
        if optimize == "UFF":
            results = AllChem.UFFOptimizeMoleculeConfs(mol, numThreads=n_threads)

    # Set energies
    if results is not None:
        energies = np.array([i[1] for i in results]) * KCAL_TO_HARTREE
    else:
        energies = None

    # Extract information from mol
    elements, conformer_coordinates, connectivity_matrix, charges = _extract_from_mol(
        mol
    )

    return (
        elements,
        conformer_coordinates,
        energies,
        connectivity_matrix,
        charges,
        mol,
    )


@requires_dependency(
    [
        Import(module="rdkit"),
        Import(module="rdkit", item="Chem"),
    ],
    globals(),
)
def _add_conformers_to_mol(mol: Chem.Mol, conformer_coordinates: ArrayLike3D) -> None:
    """Add conformers to RDKit Mol object.

    Args:
        mol: RDKit mol object
        conformer_coordinates: Conformer coordinates (Å)
    """
    conformer_coordinates: Array3DFloat = np.array(conformer_coordinates)
    if len(conformer_coordinates.shape) == 2:
        conformer_coordinates.reshape(-1, conformer_coordinates.shape[0], 3)

    for coordinates in conformer_coordinates:
        conformer = Chem.Conformer()
        for i, coord in enumerate(coordinates):
            point = rdkit.Geometry.Point3D(*coord)
            conformer.SetAtomPosition(i, point)
        mol.AddConformer(conformer, assignId=True)


@requires_dependency([Import(module="rdkit", item="Chem")], globals())
def _extract_from_mol(
    mol: Chem.Mol,
) -> tuple[list[str], Array3DFloat, Array2DFloat, Array1DInt]:
    """Extract information from RDKit Mol object with conformers."""
    # Take out elements, coordinates and connectivity matrix
    elements: list[str] = [atom.GetSymbol() for atom in mol.GetAtoms()]
    charges: Array1DInt = np.array([atom.GetFormalCharge() for atom in mol.GetAtoms()])

    conformer_coordinates = []
    for conformer in mol.GetConformers():
        coordinates = conformer.GetPositions()
        conformer_coordinates.append(coordinates)
    conformer_coordinates: Array3DFloat = np.array(conformer_coordinates)

    connectivity_matrix: Array2DFloat = Chem.GetAdjacencyMatrix(mol, useBO=True)

    return elements, conformer_coordinates, connectivity_matrix, charges


@requires_dependency(
    [
        Import(module="openbabel.openbabel", alias="ob"),
        Import(module="openbabel.pybel", alias="pybel"),
        Import("openbabel"),
    ],
    globals(),
)
def _extract_from_ob_mol(
    ob_mol: ob.OBMol,
) -> tuple[Array1DInt, Array3DFloat, Array2DInt, Array1DInt]:
    """Extract information from Openbabel OBMol object with conformers."""
    py_mol = pybel.Molecule(ob_mol)
    elements: Array1DInt = np.array([atom.atomicnum for atom in py_mol.atoms])
    charges: Array1DInt = np.array([atom.formalcharge for atom in py_mol.atoms])

    n_atoms = len(py_mol.atoms)
    connectivity_matrix: Array2DInt = np.zeros((n_atoms, n_atoms), dtype=int)
    for bond in ob.OBMolBondIter(ob_mol):
        i = bond.GetBeginAtomIdx() - 1
        j = bond.GetEndAtomIdx() - 1
        bo = bond.GetBondOrder()
        connectivity_matrix[i, j] = bo
        connectivity_matrix[j, i] = bo

    # Retrieve conformer coordinates
    conformer_coordinates = []
    for i in range(ob_mol.NumConformers()):
        ob_mol.SetConformer(i)
        coordinates: Array2DFloat = np.array([atom.coords for atom in py_mol.atoms])
        conformer_coordinates.append(coordinates)
    conformer_coordinates: Array3DFloat = np.array(conformer_coordinates)

    return elements, conformer_coordinates, connectivity_matrix, charges


@requires_dependency([Import(module="openbabel.openbabel", alias="ob")], globals())
def _get_ob_mol(
    elements: Iterable[int] | Iterable[str],
    coordinates: ArrayLike2D,
    connectivity_matrix: ArrayLike2D,
    charges: ArrayLike1D | None = None,
) -> ob.OBMol:
    """Generate OpenBabel OBMol object.

    Args:
        elements: Elements as atomic symbols or numbers.
        coordinates: Coordinates (Å)
        connectivity_matrix: Connectivity matrix with bond orders
        charges: Formal charges

    Returns:
        mol: OpenBabel OBMol object
    """
    elements = convert_elements(elements, output="numbers")

    charges_: np.ndarray
    if charges is None:
        charges_ = np.zeros(len(elements))
    else:
        charges_ = np.array(charges)
    charges = charges_

    connectivity_matrix: Array2DInt = np.array(connectivity_matrix)
    coordinates: Array2DFloat = np.array(coordinates)

    mol = ob.OBMol()

    # Add atoms
    for element, coordinate, charge in zip(elements, coordinates, charges):
        a = mol.NewAtom()
        a.SetAtomicNum(int(element))
        a.SetVector(*coordinate)
        a.SetFormalCharge(int(charge))

    # Add bonds
    for i, j in zip(*np.tril_indices_from(connectivity_matrix)):
        if i != j:
            bo = connectivity_matrix[i, j]
            if bo != 0:
                mol.AddBond(int(i + 1), int(j + 1), int(bo))
    return mol


@requires_dependency([Import(module="rdkit", item="Chem")], globals())
def _get_rdkit_mol(
    elements: Iterable[int] | Iterable[str],
    conformer_coordinates: ArrayLike3D,
    connectivity_matrix: ArrayLike2D,
    charges: ArrayLike1D | None = None,
) -> Chem.Mol:
    _RDKIT_BOND_TYPES = {
        1.0: Chem.BondType.SINGLE,
        1.5: Chem.BondType.AROMATIC,
        2.0: Chem.BondType.DOUBLE,
        3.0: Chem.BondType.TRIPLE,
        4.0: Chem.BondType.QUADRUPLE,
        5.0: Chem.BondType.QUINTUPLE,
        6.0: Chem.BondType.HEXTUPLE,
    }
    elements = convert_elements(elements, output="symbols")

    charges_: np.ndarray
    if charges is None:
        charges_ = np.zeros(len(elements))
    else:
        charges_ = np.array(charges)
    charges = charges_
    conformer_coordinates: Array3DFloat = np.array(conformer_coordinates)
    connectivity_matrix: Array2DInt = np.array(connectivity_matrix)

    mol = Chem.RWMol()

    # Add atoms
    for element, charge in zip(elements, charges):
        atom = Chem.Atom(element)
        atom.SetFormalCharge(int(charge))
        mol.AddAtom(atom)

    # Add bonds
    for i, j in zip(*np.tril_indices_from(connectivity_matrix)):
        if i != j:
            bo = connectivity_matrix[i, j]
            if bo != 0:
                bond_type = _RDKIT_BOND_TYPES[float(bo)]
                mol.AddBond(int(i), int(j), bond_type)

    # Add conformers
    _add_conformers_to_mol(mol, conformer_coordinates)

    mol = mol.GetMol()
    Chem.SanitizeMol(mol)

    return mol


def cli(smiles: str, generator: str = "rdkit") -> Any:
    """CLI for cone angle.

    Args:
        smiles: SMILES string
        generator: Confomer generator: 'ob-ff', 'ob-ga' or 'rdkit'

    Returns:
        Partially instantiated class
    """
    if generator == "rdkit":
        return functools.partial(ConformerEnsemble.from_rdkit, smiles)
    elif generator == "ob-ff":
        return functools.partial(ConformerEnsemble.from_ob_ff, smiles)
    elif generator == "ob-ga":
        return functools.partial(ConformerEnsemble.from_ob_ga, smiles)



"""Test cone angle code."""

import csv
from pathlib import Path

from numpy.testing import assert_almost_equal
import pytest

from morfeus import ConeAngle, read_xyz

DATA_DIR = Path(__file__).parent / "data" / "cone_angle"


def test_PdPMe3():
    """Test PdPMe3."""
    elements, coordinates = read_xyz(DATA_DIR / "pd/PdPMe3.xyz")
    ca = ConeAngle(elements, coordinates, 1, radii_type="bondi")
    assert_almost_equal(ca.cone_angle, 120.4, decimal=1)


def pytest_generate_tests(metafunc):
    """Generate test data from csv file."""
    if "cone_angle_data" in metafunc.fixturenames:
        with open(DATA_DIR / "cone_angles.csv") as f:
            reader = csv.DictReader(f)
            records = [("standard", record) for record in reader]
        with open(DATA_DIR / "cone_angles_max.csv") as f:
            reader = csv.DictReader(f)
            records += [("maximum", record) for record in reader]
        metafunc.parametrize("cone_angle_data", records)


@pytest.mark.benchmark
def test_reference(cone_angle_data):
    """Test against cone angle reference data."""
    for metal in ("pd", "pt", "ni"):
        label, data = cone_angle_data
        cone_angle_ref = float(data[f"{metal}_cone_angle"])
        if label == "standard":
            xyz_path = DATA_DIR / f"{metal}/{data[f'{metal}_xyz']}.xyz"
        elif label == "maximum":
            xyz_path = DATA_DIR / f"{metal}/maximum/{data[f'{metal}_xyz']}.xyz"
        elements, coordinates = read_xyz(xyz_path)
        ca = ConeAngle(elements, coordinates, 1, radii_type="bondi")
        assert_almost_equal(ca.cone_angle, cone_angle_ref, decimal=1)



"""Test XTB code."""

from pathlib import Path

import numpy as np
from numpy.testing import assert_array_almost_equal
import pytest

from morfeus import read_xyz, XTB

DATA_DIR = Path(__file__).parent / "data" / "xtb"


@pytest.mark.xtb
def test_fukui():
    """Test Fukui coefficients."""
    elements, coordinates = read_xyz(DATA_DIR / "1-penten-3-one.xyz")
    xtb = XTB(elements, coordinates)

    ref_data = np.genfromtxt(DATA_DIR / "fukui.csv", delimiter=",", names=True)
    f_nuc = list(xtb.get_fukui(variety="nucleophilicity").values())
    assert_array_almost_equal(f_nuc, ref_data["f_nuc"], decimal=3)
    f_elec = list(xtb.get_fukui(variety="electrophilicity").values())
    assert_array_almost_equal(f_elec, ref_data["f_elec"], decimal=3)
    f_rad = list(xtb.get_fukui(variety="radical").values())
    assert_array_almost_equal(f_rad, ref_data["f_rad"], decimal=3)
    f_dual = list(xtb.get_fukui(variety="dual").values())
    assert_array_almost_equal(f_dual, ref_data["f_dual"], decimal=3)
    f_loc_nuc = list(xtb.get_fukui(variety="local_nucleophilicity").values())
    assert_array_almost_equal(f_loc_nuc, ref_data["f_loc_nuc"], decimal=3)
    f_loc_elec = list(
        xtb.get_fukui(variety="local_electrophilicity", corrected=False).values()
    )
    assert_array_almost_equal(f_loc_elec, ref_data["f_loc_elec"], decimal=3)



"""Test SASA code."""

import csv
from pathlib import Path

import numpy as np
from numpy.testing import assert_almost_equal, assert_array_almost_equal
import pytest

from morfeus import read_xyz, SASA
from morfeus.typing import Array1DFloat

DATA_DIR = Path(__file__).parent / "data" / "sasa"


def test_one():
    """Test complex one."""
    atom_areas_ref = np.loadtxt(DATA_DIR / "atom_areas/1.txt")

    elements, coordinates = read_xyz(DATA_DIR / "xyz" / "1.xyz")
    sasa = SASA(elements, coordinates)
    atom_areas: Array1DFloat = np.array(list(sasa.atom_areas.values()))
    assert_almost_equal(sasa.area, 624, decimal=0)
    assert_array_almost_equal(atom_areas, atom_areas_ref, decimal=0)


def pytest_generate_tests(metafunc):
    """Generate test data from csv file."""
    if "sasa_data" in metafunc.fixturenames:
        with open(DATA_DIR / "areas.csv") as f:
            reader = csv.DictReader(f)
            records = list(reader)
            records = [(i + 1, record) for i, record in enumerate(records)]
        metafunc.parametrize("sasa_data", records)


@pytest.mark.benchmark
def test_reference(sasa_data):
    """Test against SASA reference data."""
    idx, data = sasa_data
    sasa_ref = float(data["area"])
    atom_areas_ref = np.loadtxt(DATA_DIR / f"atom_areas/{idx}.txt")
    elements, coordinates = read_xyz(DATA_DIR / "xyz" / f"{idx}.xyz")
    sasa = SASA(elements, coordinates)
    atom_areas: Array1DFloat = np.array(list(sasa.atom_areas.values()))
    assert_almost_equal(sasa.area, sasa_ref, decimal=0)
    assert_array_almost_equal(atom_areas, atom_areas_ref, decimal=0)
